! Part of TsunAWI
! Copyright (c) 2014
! Alfred Wegener Institut,
! Helmholtz Centre for Polar and Marine Research
! Contact: rakowsky@awi.de

! MODULE: MESH --  setup 2D mesh used for the tsunami simulations and the nodal properties of the mesh

MODULE MESH

  USE DATA_TYPES
  USE PARAMETERS
  USE mpi_parallel
  USE mpi_communication
  
  IMPLICIT NONE

  SAVE  


  INTEGER                                           :: elem2D
  INTEGER,   ALLOCATABLE, DIMENSION(:,:)            :: elem_nod

  INTEGER,   ALLOCATABLE, DIMENSION(:,:)            :: edge_nod
  INTEGER                                           :: edg2D
  INTEGER,   ALLOCATABLE, DIMENSION(:,:)            :: edge_elem
  INTEGER,   ALLOCATABLE, DIMENSION(:,:)            :: elem_edge
  REAL(kind=wp),      ALLOCATABLE, DIMENSION(:,:)   :: edg_nxy
  REAL(kind=wp),      ALLOCATABLE, DIMENSION(:)     :: edg_length

  INTEGER                                           :: nod2D, nbnd, nedgbnd
 
  REAL(kind=wp),      ALLOCATABLE, DIMENSION(:,:)   :: coord_nod
  real(kind=wp)                                     :: BoundingBox_xmin 
  real(kind=wp)                                     :: BoundingBox_xmax 
  real(kind=wp)                                     :: BoundingBox_ymin 
  real(kind=wp)                                     :: BoundingBox_ymax 
  INTEGER,           ALLOCATABLE, DIMENSION(:)      :: index_nod
  REAL(kind=wp),      ALLOCATABLE, DIMENSION(:)     :: cos_nod
  INTEGER,           ALLOCATABLE, DIMENSION(:)      :: nod_in_bnd, edg_in_bnd
  INTEGER,           ALLOCATABLE, DIMENSION(:,:)    :: nodngh_in_bnd
  REAL(kind=wp),      ALLOCATABLE, DIMENSION(:,:)   :: nodedgfac_in_bnd

  TYPE(addresstype), ALLOCATABLE, DIMENSION(:)      :: nghbr_nod

  REAL(kind=wp),      ALLOCATABLE, DIMENSION(:)     :: nodhn, nodhn_aux
  REAL(kind=sp),      ALLOCATABLE, DIMENSION(:)     :: nodhn_init
  REAL(kind=wp),      ALLOCATABLE, DIMENSION(:)     :: rgh_edg

! To streamline compute_velocity_at_nodes
! Compact storage for edges adjacent to nodes
  integer,           ALLOCATABLE, DIMENSION(:)      :: nodedge_ptr, nodedge_addr
  real(kind=wp),      ALLOCATABLE, DIMENSION(:)     :: nodedge_wght

! Compact storage for elements adjacent to nodes
  integer,           ALLOCATABLE, DIMENSION(:)      :: nodelem_ptr, nodelem_addr
  real(kind=wp),      ALLOCATABLE, DIMENSION(:,:)   :: nodelem_wght ! To streamline compute_ssh

  real(kind=wp)                                     :: shortestEdge, longestEdge

  CHARACTER(len=100)                                :: netcdf_file
  CHARACTER(len=100)                                :: restart_file

  ! bottom friction, variable bottom roughness
  real(KIND=wp), allocatable, dimension(:)          :: mng_edg2D


end MODULE MESH
  

SUBROUTINE mesh_scaling  

  ! Transforms degrees in rad in coord_nod(2,nod2D)     
  ! Constructs the arrays cos_nod(nod2D)

  use mesh
  IMPLICIT NONE

  INTEGER                  :: n, el
  real(kind=WP), parameter :: one_third = 1._WP/3._WP 

! Set scaling for earth or plane,
! define element wise field of mean cosine(latitude) in triangles.

    if ( coordinate_type == 1 ) then
! on the sphere 
       scaling = r_earth
       ALLOCATE(cos_nod(myDim_nod2D+eDim_nod2D))
       ! transform coordinates from deg to rad
!$OMP PARALLEL
!$OMP DO
       do n=1,myDim_nod2D+eDim_nod2D
          cos_nod(n)     = COS(coord_nod(2,n)*rad)
          coord_nod(1,n) = coord_nod(1,n) * rad
          coord_nod(2,n) = coord_nod(2,n) * rad
       END DO
!$OMP END DO
!$OMP END PARALLEL


    elseif ( coordinate_type == 2 ) then
! in the plane 
       scaling=1._wp
    else
       print *,'Sorry, there is no coordinate_type =',coordinate_type,'implemented'
       stop 1
    end if

END SUBROUTINE mesh_scaling


SUBROUTINE read_2Dmesh()
  !
  !--------------------------------------------------------------------------
  ! READS 2D MESH AND VARIABLES CONCERNING THE 2D MESH    
  !
  ! OUTPUT:
  !   Number of 2D-Nodes: nod2D
  !   Number of 2D-Elements: elem2D
  !   Number of 2D-element-neighbours is always 3
  !
  !   Mesh-Arrays: 
  !     coord_nod(2,nod2D)
  !     elem_nod(3,elem2D)
  !     index_nod(nod2D)
  !     nodhn(nod2D)
  !--------------------------------------------------------------------------
  !
!$  use OMP_LIB
  use mesh
  IMPLICIT NONE
  !
  INTEGER            :: infile_nod, infile_elem, infile_depth
  INTEGER            :: nq, ind
  REAL(kind=wp)      :: x1, x2
  INTEGER            :: n, i_err, el
  character(len=200) :: a_dummy

#ifdef USE_MPI     

  logical                    :: found
  real(kind=wp), allocatable :: coord_glob(:,:), nodhn_glob(:)
  integer, allocatable       :: index_glob(:)
  INTEGER                    :: i, n_glob, n_el

  ! Version for distributed subdomains - complicated! ;-)

  if (npes > 1) then
     ! First, each PE reads it's own distribution
     call com_read_subdomains(MeshPath)

     allocate(elem_nod(3, myDim_elem2D+eDim_elem2D+eXDim_elem2D)) 
!$OMP PARALLEL DO  
!$  DO el=1, myDim_elem2D+eDim_elem2D+exDim_elem2D
!$     elem_nod(:,el) = 0
!$  END DO
!$OMP END PARALLEL DO


     call com_read_elements_on_subdomains(MeshPath, elem_nod)
     
     if (mype==0) then
        OPEN(newunit=infile_nod,   file= TRIM(MeshPath) // 'nod2d.out',  status= 'old')
        WRITE(*,*) '2D mesh is opened'
        
        READ(infile_nod, *) nod2D       
        
        !NR I would prefer to avoid ifdef here... but this would need
        !NR some restructuring.
#ifdef USE_MPI     
        com_nod2D%globalDim       = nod2D
#endif     
        
        ALLOCATE(coord_glob(2,nod2D), index_glob(nod2D))
        
        ! Test, if nod2d.out contains 3 or 4 columns.
        ! The first column, the number of the node, may be skipped.
        READ(infile_nod,'(a)') a_dummy
        READ(a_dummy,*,IOSTAT=i_err) nq, x1, x2, ind
        
        if (i_err==0) then
           coord_glob(1,nq) = x1
           coord_glob(2,nq) = x2
           index_glob(nq)   = ind
           DO n=2, nod2D
              READ(infile_nod, *) nq, x1, x2, ind
              coord_glob(1,nq) = x1
              coord_glob(2,nq) = x2
              index_glob(nq)   = ind
           END DO
        else
           READ(a_dummy,*) coord_glob(1:2,1),  index_glob(1)
           DO n=2, nod2D
              READ(infile_nod, *) coord_glob(1:2,n),  index_glob(n)
           end DO
        end if
        CLOSE(infile_nod)
  ! Determine Bounding Box for later use

        BoundingBox_xmin = coord_glob(1,1)
        BoundingBox_xmax = coord_glob(1,1)
        BoundingBox_ymin = coord_glob(2,1)
        BoundingBox_ymax = coord_glob(2,1)
!$OMP PARALLEL DO &
!$OMP       reduction(min: BoundingBox_xmin, BoundingBox_ymin) &  
!$OMP       reduction(max: BoundingBox_xmax, BoundingBox_ymax)  
        do n=2,nod2D
           BoundingBox_xmin = min(BoundingBox_xmin,coord_glob(1,n))
           BoundingBox_xmax = max(BoundingBox_xmax,coord_glob(1,n))
           BoundingBox_ymin = min(BoundingBox_ymin,coord_glob(2,n))
           BoundingBox_ymax = max(BoundingBox_ymax,coord_glob(2,n))
        end do
!$OMP END PARALLEL DO

     else ! mype /=0

        ! dummy allocation
        allocate(coord_glob(1,1))
        allocate(index_glob(1))
     endif

  ! For OpenMP, first touch for data locality
     allocate(coord_nod(2,myDim_nod2D + eDim_nod2D))
     allocate(index_nod(  myDim_nod2D + eDim_nod2D))
!$OMP PARALLEL DO
!$ DO n = 1, myDim_nod2D + eDim_nod2D
!$   coord_nod(1:2,n) = 0._wp
!$   index_nod(n)     = 0  
!$ END DO  
!$OMP END PARALLEL DO

     call com_scatter(com_nod2D, coord_glob, coord_nod)
     call com_scatter(com_nod2D, index_glob, index_nod)

     call com_halo_begin(com_nod2D, coord_nod)
     call com_halo_begin(com_nod2D, index_nod)  
     deallocate(coord_glob, index_glob)
     call com_halo_end(com_nod2D)

  
     !
     !  READS THE NODE-NUMBERS OF EACH ELEMENT 
     !
     ! We need elem2D for IO - get this later by other means...
     OPEN(newunit=infile_elem,  file= TRIM(MeshPath) // 'elem2d.out', status= 'old')
     read(infile_elem, *) elem2D
#ifdef USE_MPI     
     com_elem2D%globalDim      = elem2D
     com_elem2D_full%globalDim = elem2D
#endif
     close(infile_elem)
     

  
  
  ! And the bathymetry

     if (mype==0) then
        ALLOCATE(nodhn_glob(nod2D))
        
        OPEN(newunit=infile_depth, file= TRIM(MeshPath)//TRIM(TopoFile),  status= 'old')  
        READ(infile_depth, *) nodhn_glob(1:nod2D)
        CLOSE(infile_depth)
     else
        allocate(nodhn_glob(1))
     end if
     
     ALLOCATE(nodhn(     myDim_nod2D + eDim_nod2D))
!$OMP PARALLEL DO
!$  DO n=1,myDim_nod2D+eDim_nod2D
!$     nodhn(n) = 0._wp
!$  END DO
!$OMP END PARALLEL DO

     call com_scatter(com_nod2D, nodhn_glob, nodhn)
     call com_halo_begin(com_nod2D, nodhn)
     call com_halo_end(com_nod2D)

     deallocate(nodhn_glob)

     
!=========================================================     
  else   ! n_pes=1  or   no MPI
!=========================================================     
#endif
     
     ! The simple case. No distributed mesh, just read!
     OPEN(newunit=infile_nod,   file= TRIM(MeshPath) // 'nod2d.out',  status= 'old')
     WRITE(*,*) '2D mesh is opened'
        
     READ(infile_nod, *) nod2D       

     ALLOCATE(coord_nod(2,nod2D), index_nod(nod2D))
       
     ! For OpenMP, first touch for data locality
!$OMP PARALLEL DO
!$ DO n = 1, nod2D
!$   coord_nod(1:2,n) = 0._wp
!$   index_nod(n)     = 0  
!$ END DO  
!$OMP END PARALLEL DO
 
     ! Test, if nod2d.out contains 3 or 4 columns.
     ! The first column, the number of the node, may be skipped.
     READ(infile_nod,'(a)') a_dummy
     READ(a_dummy,*,IOSTAT=i_err) nq, x1, x2, ind
        
     if (i_err==0) then
        coord_nod(1,nq) = x1
        coord_nod(2,nq) = x2
        index_nod(nq)   = ind
        DO n=2, nod2D
           READ(infile_nod, *) nq, x1, x2, ind
           coord_nod(1,nq) = x1
           coord_nod(2,nq) = x2
           index_nod(nq)   = ind
        END DO
     else
        READ(a_dummy,*) coord_nod(1:2,1),  index_nod(1)
        DO n=2, nod2D
           READ(infile_nod, *) coord_nod(1:2,n),  index_nod(n)
        end DO
     end if
     CLOSE(infile_nod)
  ! Determine Bounding Box for later use

     BoundingBox_xmin = coord_nod(1,1)
     BoundingBox_xmax = coord_nod(1,1)
     BoundingBox_ymin = coord_nod(2,1)
     BoundingBox_ymax = coord_nod(2,1)
!$OMP PARALLEL DO &
!$OMP       reduction(min: BoundingBox_xmin, BoundingBox_ymin) &  
!$OMP       reduction(max: BoundingBox_xmax, BoundingBox_ymax)  
     do n=2,nod2D
        BoundingBox_xmin = min(BoundingBox_xmin,coord_nod(1,n))
        BoundingBox_xmax = max(BoundingBox_xmax,coord_nod(1,n))
        BoundingBox_ymin = min(BoundingBox_ymin,coord_nod(2,n))
        BoundingBox_ymax = max(BoundingBox_ymax,coord_nod(2,n))
     end do
!$OMP END PARALLEL DO  

     !  READS THE NODE-NUMBERS OF EACH ELEMENT 
  !
     OPEN(newunit=infile_elem,  file= TRIM(MeshPath) // 'elem2d.out', status= 'old')
     READ(infile_elem, *)  elem2D
        
     ALLOCATE(elem_nod(3,elem2D)) 
!$OMP PARALLEL DO  
!$  DO el=1, elem2D
!$     elem_nod(:,el) = 0
!$  END DO
!$OMP END PARALLEL DO
        
     read(infile_elem, *) elem_nod(1:3,1:elem2D)
     CLOSE(infile_elem)
  
  ! And the bathymetry

     ALLOCATE(nodhn(nod2D))
!$OMP PARALLEL DO
!$  DO n=1,nod2D
!$     nodhn(n) = 0._wp
!$  END DO
!$OMP END PARALLEL DO
        
     OPEN(newunit=infile_depth, file= TRIM(MeshPath)//TRIM(TopoFile),  status= 'old')  
     READ(infile_depth, *) nodhn(1:nod2D)
     CLOSE(infile_depth)

     ! Set the MPI domain decomposition style array bounds:
     myDim_nod2D  = nod2D
     eDim_nod2D   = 0
     myDim_elem2D = elem2D
     eDim_elem2D  = 0
     exDim_elem2D = 0
     
#ifdef USE_MPI 
  endif  ! MPI with npes > 1
#endif  

  ! The initial topography is kept for reference, e.g. to determine land nodes
  ! before the rupture lifts or lowers nodes. The main purpose is to retain the
  ! original land-sea-mask for output.
  ! Furthermore, nodhn will be smoothed.
  
  ALLOCATE(nodhn_init(myDim_nod2D))
!$OMP PARALLEL DO
  DO n=1,myDim_nod2D
     nodhn_init(n) = real(nodhn(n), sp)
  END DO
!$OMP END PARALLEL DO
     
END SUBROUTINE read_2Dmesh

!--------- --------- --------- --------- --------- --------- ---------

SUBROUTINE set_up_bv_roughness

  use mesh
  IMPLICIT NONE

  integer                          :: edg, nd1, nd2, n
  REAL(kind=wp), DIMENSION(myDim_nod2D+eDim_nod2D)  :: rgh_nod
  integer                          :: infile_rgh

  OPEN(newunit=infile_rgh,file= TRIM(MeshPath) // 'rgh.out',  status= 'old')
  DO n=1, myDim_nod2D
     read(infile_rgh,*) rgh_nod(n)
  END DO
  CLOSE(infile_rgh)     
  

  ALLOCATE(rgh_edg(myDim_edg2D))
  
!$OMP PARALLEL DO 
  do edg=1,myDim_edg2D

     nd1=edge_nod(1,edg)
     nd2=edge_nod(2,edg)

     if ( nodhn(nd1)<150._wp .OR. nodhn(nd2)<150._wp ) then
        rgh_edg(edg)=0. 
     else
        rgh_edg(edg)=0.5_wp*(rgh_nod(nd1)+rgh_nod(nd2))
     endif

  end do
!$OMP END PARALLEL DO

END SUBROUTINE set_up_bv_roughness



subroutine set_up_mng_roughness

  use mesh
  implicit none

  integer                         :: i, n, nmb
  integer                         :: edg
  real(kind=wp), dimension(myDim_nod2D+eDim_nod2D) :: mng_aux, mng_nod
  integer                         :: infile_mng

  print *,'Setting up Manning roughness'

!SH Manning roughness field

    open(newunit=infile_mng, file= trim(MeshPath)//trim(BottomFrictionFile), status= 'old')
    read(infile_mng,*) mng_nod(1:nod2D)
    
    close(infile_mng)

if (mng_max > 0.) then
!$OMP PARALLEL DO
  do n=1,myDim_nod2D
     mng_nod(n) = min(max(mng_nod(n),mng_min),  mng_max) 
  end do
!$OMP END PARALLEL DO
else  
!$OMP PARALLEL DO
  do n=1,myDim_nod2D
     mng_nod(n) = max(mng_nod(n),mng_min) 
  end do
!$OMP END PARALLEL DO
endif

  print *,'Before smoothing:'
  print *,'Minimal Manning roughness : ',minval(mng_nod)
  print *,'Maximal Manning roughness : ',maxval(mng_nod)

  do i = 1,2  
!$OMP PARALLEL 
!$OMP DO  private(nmb)
    do n = 1, myDim_nod2D
      nmb = nghbr_nod(n)%nmb
      mng_aux(n) = sum(mng_nod(nghbr_nod(n)%addr(:))) / real(nmb,wp)
    end do
!$OMP END DO 
!$OMP DO 
    do n = 1, myDim_nod2D
       mng_nod(n)=mng_aux(n)
    end do
!$OMP END DO 
!$OMP END PARALLEL
  end do

  print *,'After smoothing:'
  print *,'Minimal Manning roughness : ',minval(mng_nod)
  print *,'Maximal Manning roughness : ',maxval(mng_nod)
  

  allocate(mng_edg2D(myDim_edg2D))

!$OMP PARALLEL DO 
  do edg=1,myDim_edg2D
    mng_edg2D(edg)=0.5_wp*sum(mng_nod(edge_nod(1:2,edg)))
  end do  
!$OMP END PARALLEL DO

  print *,'Values in edges :'
  print *,'Minimal Manning roughness : ',minval(mng_edg2D)
  print *,'Maximal Manning roughness : ',maxval(mng_edg2D)



end subroutine set_up_mng_roughness


SUBROUTINE build_nghbr_arrays()
  !
  use mesh
  
  IMPLICIT NONE
  
  INTEGER                 :: j,k,m,a,b,c, count, el,ml(1)
  INTEGER, DIMENSION(100) :: aux=0
  INTEGER                 :: ind(myDim_nod2D+eDim_nod2D)
  integer                 :: n, iA, iZ
  integer                 :: outfile_nghb
  !
  !--------------- 2D mesh:
  ! Builds sparse storage for patches: nodelem_ptr, nodelem_addr
  !   

  ind(:) = 0
  DO el = 1, myDim_elem2D 
     ind(elem_nod(1:3,el)) = ind(elem_nod(1:3,el)) + 1
  END DO
  ! The halo elements need special treatment, as the outer nodes do not
  ! belong to the local submesh.
  ! This outer nodes are indicates with their negative global numbering.
  ! Thus, collect only the nodes with positive index -> local nodes
  DO el = myDim_elem2D+1,  myDim_elem2D+eDim_elem2D+exDim_elem2D
     if (elem_nod(1,el) > 0) ind(elem_nod(1,el)) = ind(elem_nod(1,el)) + 1
     if (elem_nod(2,el) > 0) ind(elem_nod(2,el)) = ind(elem_nod(2,el)) + 1
     if (elem_nod(3,el) > 0) ind(elem_nod(3,el)) = ind(elem_nod(3,el)) + 1
  END DO

  allocate(nodelem_ptr(myDim_nod2D+eDim_nod2D+1))
! first touch initialization for data locality in OpenMP
!$OMP PARALLEL DO 
!$  DO n = 1, myDim_nod2D+eDim_nod2D+1
!$     nodelem_ptr(n) = 0
!$  END DO
!$OMP END PARALLEL DO

  nodelem_ptr(1) = 1

  DO n = 1, myDim_nod2D+eDim_nod2D
     nodelem_ptr(n+1) = nodelem_ptr(n) + ind(n)
  END DO

  allocate(nodelem_addr(nodelem_ptr(myDim_nod2D+eDim_nod2D+1)-1))

!$OMP PARALLEL DO  PRIVATE(n,j) 
  DO n = 1, myDim_nod2D+eDim_nod2D
     do j = nodelem_ptr(n), nodelem_ptr(n+1)-1
        nodelem_addr(j) = 0
     enddo
  END DO
!$OMP END PARALLEL DO      

  DO el = 1, myDim_elem2D + eDim_elem2D + eXDim_elem2D
     DO k = 1, 3  
        a = elem_nod(k,el)
        if (a>0 .and. a <= myDim_nod2D+eDim_nod2D) then
           jLoop: do j = nodelem_ptr(a), nodelem_ptr(a+1)-1
              if (nodelem_addr(j) == 0 ) then
                 nodelem_addr(j) = el
                 exit jLoop
              endif
           enddo jLoop
        endif
     end DO
  END DO


  ! The list of elements is ordered, and no sorting is needed
  !
 
  ALLOCATE(nghbr_nod(myDim_nod2D+eDim_nod2D))
  
  DO n = 1, myDim_nod2D+eDim_nod2D
     count = 1
     aux(1) = n
     DO m = nodelem_ptr(n), nodelem_ptr(n+1)-1 
        el = nodelem_addr(m)
        DO k = 1, 3
           a = elem_nod(k,el)       
           IF(a/=n .and. all(aux(1:count) /= a)) THEN  
              count = count + 1         
              aux(count) = a
           END IF
        END DO
     END DO
     nghbr_nod(n)%nmb = count
     ALLOCATE(nghbr_nod(n)%addr(count))
     
     ! we need to sort array aux(1:count)
     DO m = count, 1, -1
        ml = MAXLOC(aux(1:count))
        nghbr_nod(n)%addr(m) = aux(ml(1))
        aux(ml(1)) = -999
     END DO
  END DO

    ! perform auxiliary output
  if (write_ascii_neighbourhood) then
     if (npes==1) then
        open(newunit=outfile_nghb, file="neighbourhood.out")
        do n=1,myDim_nod2D+eDim_nod2D
           WRITE(outfile_nghb,*) nghbr_nod(n)%nmb
           do j=1,nghbr_nod(n)%nmb
              WRITE(outfile_nghb,*) nghbr_nod(n)%addr(j)
           end do
        end do
        close(outfile_nghb)
     else
        if (mype==0) then
           print *,'!!! write_ascii_neighbourhood is not implemented for MPI distributed mesh!!!'
           print *,'Please start TsunAWI with only one MPI task.'
        endif
     end if
  end if

END SUBROUTINE build_nghbr_arrays


SUBROUTINE build_edg_arrays
  ! build_edg_arrays
  !OUTPUT:
  !edg2D
  !edge_nod     (2, edg2D)                       
  !edge_elem (2, edg2D)
  !edg_nxy        (2, edg2D)
  !edg_length     (   edg2D)
  !elem_edge   (3, elem2D)
  !--------------------------------------------------------------------------
  !edg2D
  !the total number of edges
  !
  !edge_nod(:, edg) 
  !keeps two nodes which form the edge edg. edge_nod(edg,2)>edge_nod(edg, 1). 
  !
  !elem_edge(:, elem)
  !keeps the  edges which belong to the elem. (/1, 2/), (/2, 3/), (/3, 1/) are 
  !the nodes (elem_nod(:, elem)) of a triangle elem which form these edges.
  !
  !edge_elem(:, edg)
  !keeps two triangle which have the common edge edg. If it is only one triangle 
  !the second value will be zero. edge_elem(1, edg) is on the left side of 
  !a direction (edge_nod(1, edg) edge_nod(2, edg)). 
  !
  !edg_nxy(:, edg)
  !is the normal (nx, ny) to the edge always points from the left triangle to the 
  !right one. If the edge belongs to the open boundary than it will be the outer 
  !normal to the edge.
  !
  !edg_length(edg)
  ! length of the edge [m]
  !

  use mesh
  IMPLICIT NONE
  !
  INTEGER            :: n, el, cnt, n1
  INTEGER            :: i, j, k, elnodes(3), iA, iZ, edg
  REAL(kind=wp)      :: vec(2), vec2(2), r, lvec
  real(kind=wp)      :: one_third = 1._wp/3._wp
  
  !NR We do have
  !NR  - the inner edges, both nodes <= myDim_nod2D
  !NR  - the edges inside the halo, one node <= myDim_nod2D, the other > myDim_nod2D
  !NR  - the edges at rim of the halo, both nodes > myDim_nod2D

  
  myDim_edg2D = 0
!$OMP PARALLEL DO REDUCTION(+:myDim_edg2D)  
  DO n = 1, myDim_nod2D+eDim_nod2D
     myDim_edg2D = myDim_edg2D + COUNT(nghbr_nod(n)%addr(1:nghbr_nod(n)%nmb) > n)
  END DO
!$OMP END PARALLEL DO
  edg2D = myDim_edg2D
  
  if (verbosity >= 5) then 
     print *,'Number of nodes    in 2D mesh: ', myDim_nod2D,'+',eDim_nod2D
     print *,'          elements in 2D mesh: ', myDim_elem2D,'+',eDim_elem2D,'+',exDim_elem2D
     print *,'          edges    in 2D mesh: ', myDim_edg2D
  endif

  ALLOCATE(edge_nod( 2, myDim_edg2D))
  ALLOCATE(edge_elem(2, myDim_edg2D))
  ALLOCATE(elem_edge(3, myDim_elem2D+eDim_elem2D+exDim_elem2D))
  ALLOCATE(edg_nxy(  2, myDim_edg2D))
  ALLOCATE(edg_length(  myDim_edg2D))
!$OMP PARALLEL 
!$OMP DO 
  do n=1,  myDim_edg2D
     edge_nod(:,n)  = 0
     edge_elem(:,n) = 0
  enddo
!$OMP END DO NOWAIT
!$OMP DO 
  do n=1, myDim_elem2D+eDim_elem2D+exDim_elem2D
     elem_edge(:,n)   = 0
  enddo
!$OMP END DO
!$OMP END PARALLEL 

  cnt = 0 
  DO n = 1, myDim_nod2D+eDim_nod2D
     DO i = 1, nghbr_nod(n)%nmb
        IF (nghbr_nod(n)%addr(i) > n) then
           cnt = cnt+1
           edge_nod(:,cnt) = (/ n, nghbr_nod(n)%addr(i) /)
        endif
     end DO
  end DO

  
  vec(1:2) = coord_nod(:,1) - coord_nod(:,2)
  shortestEdge = SQRT(SUM(vec*vec)) * scaling
  longestEdge  = 0.
  
!$OMP PARALLEL DO DEFAULT(SHARED) &
!$OMP&         PRIVATE(n,el,n1, j, edg,vec,vec2,lvec) &  
!$OMP&         REDUCTION(min:shortestEdge) &
!$OMP&         REDUCTION(max: longestEdge) 
  do edg = 1, myDim_edg2D
     n = edge_nod(1,edg)
     n1 = edge_nod(2,edg)

     vec(1:2) = coord_nod(:,n1) - coord_nod(:,n)
     
     if (coordinate_type == 1) then
        !correct global grids for jumps across -180/180
        if (abs(vec(1)) > .5_wp*pi) vec(1) = -sign(1._wp,vec(1))*(2._wp*pi-abs(vec(1)))
        ! and scale according to latitude
        vec(1) = vec(1) * .5_WP*(cos_nod(n)+cos_nod(n1))
     endif

     DO j = nodelem_ptr(n),nodelem_ptr(n+1)-1 
        el = nodelem_addr(j) 
        !
        !Finds whether edge edg belongs to the element el or not
        !
        IF (any(elem_nod(:,el) == n1)) THEN

           if (all(elem_nod(:,el) > 0)) then
              ! simple case, it is an inner element of our subdomain
              ! and we have the full information on this element

              
              if (elem_nod(1,el) /= n .and. elem_nod(1,el) /= n1) then
                 vec2(1:2) = coord_nod(:,elem_nod(1,el)) - coord_nod(:,n)
                 elem_edge(2, el) = edg
              elseif (elem_nod(2,el) /= n .and. elem_nod(2,el) /= n1) then
                 vec2(1:2) = coord_nod(:,elem_nod(2,el)) - coord_nod(:,n)
                 elem_edge(3, el) = edg
              else
                 vec2(1:2) = coord_nod(:,elem_nod(3,el)) - coord_nod(:,n)
                 elem_edge(1, el) = edg
              END IF

              
              if (coordinate_type == 1) then
                 !correct for jumps across -180/180 on global meshes
                 if (abs(vec2(1)) > .5_wp*pi) &
                      vec2(1) = -sign(1._wp,vec2(1))*(2._wp*pi-abs(vec2(1)))
                 ! and scale according to latitude (approximately, is sufficient here)
                 vec2(1) = vec2(1) * .5_WP*(cos_nod(n)+cos_nod(n1))
              endif

              !Finds whether element el is on the left or on the right of vec
              !
              IF (vec(1) * vec2(2) - vec(2) * vec2(1) > 0) THEN
                 edge_elem(1,edg) = el
              ELSE
                 ! The first element was placed in the wrong location - correct here.
                 ! This can only happen for halo elements in a domain decomposition.
                 if (edge_elem(2,edg) > 0) edge_elem(1,edg)=edge_elem(2,edg)
                 edge_elem(2,edg) = el
              END IF
              
            
           else ! This can only happen for npes > 1, element is a halo element
              ! and we do not own the third node of this element

              ! We drop this element as the second neighbour as default. If this should
              ! be wrong, it will be corrected when the other neighbour (inner element)
              ! is found and evaluated.

              if (edge_elem(2,edg) > 0) then
                 edge_elem(1,edg) = el
              else
                 edge_elem(2,edg) = el
              endif
                 
              ! This elem has only one known edge
              ! However, the ordering is still important - later, we have
              ! to choose the right basis function for the edge!
                        
              if (elem_nod(1,el) /= n .and. elem_nod(1,el) /= n1) then
                 elem_edge(2,el) = edg
              elseif (elem_nod(2,el) /= n .and. elem_nod(2,el) /= n1) then
                 elem_edge(3,el) = edg
              else
                 elem_edge(1,el) = edg
              END IF
              
           end if
        END IF
     END DO
     !
     !Now construct the normal to the edge 
     !
     lvec = SQRT(SUM(vec*vec))
     
     edg_nxy(1,edg) =  vec(2) / lvec
     edg_nxy(2,edg) = -vec(1) / lvec     
     edg_length(edg) = lvec * scaling

     shortestEdge = min(edg_length(edg),shortestEdge)
     longestEdge  = max(edg_length(edg), longestEdge)
     
     IF (edge_elem(1,edg) == 0) THEN
        edg_nxy(1,edg) = -edg_nxy(1,edg)
        edg_nxy(2,edg) = -edg_nxy(2,edg)
        edge_elem(1,edg) = edge_elem(2,edg)
        edge_elem(2,edg) = 0
        edge_nod(:,edg) = (/ edge_nod(2,edg), edge_nod(1,edg) /)
     END IF
  END DO
!$OMP END PARALLEL DO

  ! pe 0 needs the global min / max edge length for output
  call com_reduce_begin(shortestEdge, com_min)
  call com_reduce_begin( longestEdge, com_max)
  
!  if (allocated(cos_nod)) deallocate(cos_nod)

! Build compact storage: node with surrounding edges 
  allocate(nodedge_ptr(myDim_nod2D+eDim_nod2D+1))
! first touch initialization for data locality in OpenMP
!$OMP PARALLEL DO 
!$  DO n = 1, myDim_nod2D+1
!$     nodedge_ptr(n) = 0
!$  END DO
!$OMP END PARALLEL DO

  nodedge_ptr(1) = 1
  do n=1,myDim_nod2D+eDim_nod2D
    nodedge_ptr(n+1) = nodedge_ptr(n) +  nghbr_nod(n)%nmb -1
 enddo

 allocate(nodedge_addr(nodedge_ptr(myDim_nod2D+eDim_nod2D+1)-1))

! first touch initialization for data locality in OpenMP 
!$OMP PARALLEL DO  PRIVATE(n,j) 
  DO n = 1, myDim_nod2D+eDim_nod2D
     do j = nodedge_ptr(n), nodedge_ptr(n+1)-1
        nodedge_addr(j) = 0
     enddo
  END DO
!$OMP END PARALLEL DO      


 do edg=1,myDim_edg2D

! insert edg as adjacent to it's 1st and 2nd node
    do j=1,2
       n = edge_nod(j,edg)
       if (n <= myDim_nod2D) then
          iA = nodedge_ptr(n)
          iZ = nodedge_ptr(n+1)-1
          do i=iA,iZ
             if (nodedge_addr(i)==0) then
                nodedge_addr(i) = edg
                exit
             endif
          enddo
       endif
    enddo ! j=1,2

 end do ! edg

 call com_reduce_end
 
END SUBROUTINE build_edg_arrays

SUBROUTINE smooth_topography

  use mesh
  
  INTEGER                 :: i, n, ierr, req(3)
  REAL(kind=wp)           :: depval, topo_min
  REAL(kind=sp)           :: nodhn_min, nodhn_max  


  !============== PREPARING TOPOGRAPHY =========


  !setting topography in coastal nodes to zero
  i = 0
  nodhn_min = 99999.
  nodhn_max =-99999.
!$OMP PARALLEL DO  REDUCTION(+:i) &
!$OMP             REDUCTION(MIN:nodhn_min) &
!$OMP             REDUCTION(MAX:nodhn_max) 
  DO n = 1, myDim_nod2D+eDim_nod2D
     IF ( index_nod(n) == 3 .AND. nodhn(n) /= 0.0 ) THEN
        nodhn(n) = 0.0
        i = i + 1
     END IF
     nodhn_min = min(nodhn_min,nodhn(n))
     nodhn_max = max(nodhn_max,nodhn(n))
  END DO
!$OMP END PARALLEL DO
  if (verbosity > 1) then
     call com_reduce_begin(i,         com_sum)
     call com_reduce_begin(nodhn_min, com_min)
     call com_reduce_begin(nodhn_max, com_max)
     call com_reduce_end     
     if (mype==0) PRINT *,'Set topograhy in',i,' coastal nodes to Zero'
  endif
  ! smoothing	the topography 

  topo_min=-.1_wp   ! coastal land nodes should remain _land_ 
  if (enable_benchmark) topo_min=0._wp

  if (nmb_iter_smooth_topo>0) then
     allocate(nodhn_aux(myDim_nod2D+eDim_nod2D))
     if (verbosity>1 .and. mype==0) &
          print *,'Topography before smoothing: ',nodhn_min, nodhn_max

     do i = 1, nmb_iter_smooth_topo
!$OMP PARALLEL DEFAULT(SHARED) PRIVATE(n,depval)
!$OMP DO  
        do n = 1, myDim_nod2D

           !Check for (seldom) case of degenerated grid: node without neighbours
           if (nghbr_nod(n)%nmb > 0 ) then
              depval = sum(nodhn(nghbr_nod(n)%addr(:))) / real(nghbr_nod(n)%nmb,wp)

              if ( all(nodhn(nghbr_nod(n)%addr(:)) >= 0.) .OR.   &
                   all(nodhn(nghbr_nod(n)%addr(:)) <= 0.) ) then

                 nodhn_aux(n) = depval

              else
                 if (nodhn(n) >= 0.) then
                    nodhn_aux(n) = max(depval,0._wp)
                 else
                    nodhn_aux(n) = min(depval,topo_min)
                 end if
              end if

           else
              nodhn_aux(n) = nodhn(n)
           endif

        end do
!$OMP END DO        
!$OMP DO 
        do n = 1, myDim_nod2D
           nodhn(n) = nodhn_aux(n)
        end do
!$OMP END DO          
!$OMP END PARALLEL
        call com_halo_begin(com_nod2D, nodhn)
        call com_halo_end(com_nod2D)
     end do
     deallocate(nodhn_aux)


     !setting topography in coastal nodes *AGAIN* to zero
     i = 0
     nodhn_min = 99999.
     nodhn_max =-99999.
!$OMP PARALLEL DO REDUCTION(MIN:nodhn_min) &
!$OMP             REDUCTION(MAX:nodhn_max) &
!$OMP             REDUCTION(+:i) 
     do n = 1, myDim_nod2D+eDim_nod2D
        if ( index_nod(n) == 3 .and. nodhn(n) /= 0.0 ) then
           nodhn(n) = 0.0
           i = i + 1
        end if
        nodhn_min = min(nodhn_min,nodhn(n))
        nodhn_max = max(nodhn_max,nodhn(n))
     end do
!$OMP END PARALLEL DO
     
     if(verbosity > 1) then

        call com_reduce_begin(i,         com_sum)
        call com_reduce_begin(nodhn_min, com_min)
        call com_reduce_begin(nodhn_max, com_max)
        call com_reduce_end 
        if (mype==0) then
           print *,'topography after smoothing : ',nodhn_min, nodhn_max
           PRINT *,'Set topograhy in',i,' coastal nodes to Zero again'
        endif
     endif
  end if

END SUBROUTINE smooth_topography

