! ========================================================================
! Halo exchange routines and
! gather routines that collect information on the entire field (needed for output)
! ========================================================================

module mpi_communication

    USE parameters, only: wp, sp, dp  ! floating point precision
    USE mpi_parallel
    implicit none

    private 
    public :: com_halo_begin, com_halo_end
    public :: com_gather, com_scatter
    public :: com_reduce_begin, com_allreduce_begin, com_reduce_end
    public :: com_sum, com_max, com_min

    interface com_halo_begin
       module procedure com_halo_1_begin
       module procedure com_halo_n_begin
       module procedure com_halo_i_1_begin
       module procedure com_halo_l_1_begin
       module procedure com_halo_i_n_begin
    end interface com_halo_begin


    interface com_gather
       module procedure com_gather_double2double_1
       module procedure com_gather_double2double_n
       module procedure com_gather_i_1
       module procedure com_gather_i_n
       module procedure com_gather_double2single_1
       module procedure com_gather_double2single_n
       module procedure com_gather_single2single_1
       module procedure com_gather_single2single_n
    end interface com_gather
    
    interface com_scatter
       module procedure com_scatter_1
       module procedure com_scatter_n
       module procedure com_scatter_i_1
       module procedure com_scatter_i_n
    end interface com_scatter
    
    interface com_reduce_begin
       module procedure com_reduce_i_1_begin
       module procedure com_reduce_sp_1_begin
       module procedure com_reduce_dp_1_begin
       module procedure com_reduce_i_begin
       module procedure com_reduce_sp_begin
       module procedure com_reduce_dp_begin
    end interface com_reduce_begin

    interface com_allreduce_begin
       module procedure com_allreduce_i_1_begin
       module procedure com_allreduce_sp_1_begin
       module procedure com_allreduce_dp_1_begin
       module procedure com_allreduce_i_begin
       module procedure com_allreduce_sp_begin
       module procedure com_allreduce_dp_begin
    end interface com_allreduce_begin

    
    
#ifdef USE_MPI
    include 'mpif.h'

    integer, parameter :: com_sum = MPI_SUM
    integer, parameter :: com_max = MPI_MAX
    integer, parameter :: com_min = MPI_MIN
#else
    integer, parameter :: com_sum = 0
    integer, parameter :: com_max = 1
    integer, parameter :: com_min = 2
#endif
    
contains

#ifdef USE_MPI
  ! Routines for MPI communication. Dummy routines for no MPI follow below.
subroutine com_halo_i_1_begin(com, array2D)
  use omp_lib
  ! General version of the communication routine for 2D nodal fields
  type(com_struct), intent(inout) :: com     
  integer, intent(inout)       :: array2D(:)
  integer                      :: n, sn, rn, ierr, nreq

  if (npes > 1) then
     sn   = com%sPEnum
     rn   = com%rPEnum
     nreq = com%nreq

     com%nreq = com%nreq + rn+sn

     ! Expensive, but sorted out after the first timestep
     if (com%nreq > size(com%req)) call enlarge_array(com%nreq, com%req)
     
     DO n=1,rn    

        call MPI_IRECV(array2D, 1, com%rtype_i(n,1), com%rPE(n), &
             com%rPE(n), mpi_comm_tsuna, com%req(nreq + n), ierr) 
     END DO

     DO n=1, sn

        call MPI_ISEND(array2D, 1, com%stype_i(n,1), com%sPE(n), &
             mype, mpi_comm_tsuna, com%req(nreq + rn+n), ierr)
     END DO
  endif
END SUBROUTINE com_halo_i_1_begin

!=====================================================================

subroutine com_halo_l_1_begin(com, array2D)

  ! General version of the communication routine for 2D nodal fields
  type(com_struct), intent(inout) :: com     
  logical, intent(inout)          :: array2D(:)
  integer                         :: n, sn, rn, ierr, nreq

  if (npes > 1) then
     sn   = com%sPEnum
     rn   = com%rPEnum
     nreq = com%nreq

     com%nreq = com%nreq + rn+sn

     ! Expensive, but sorted out after the first timestep
     if (com%nreq > size(com%req)) call enlarge_array(com%nreq, com%req)
     
     DO n=1,rn    

        call MPI_IRECV(array2D, 1, com%rtype_l(n), com%rPE(n), &
             com%rPE(n), mpi_comm_tsuna, com%req(nreq + n), ierr) 
     END DO

     DO n=1, sn

        call MPI_ISEND(array2D, 1, com%stype_l(n), com%sPE(n), &
             mype, mpi_comm_tsuna, com%req(nreq + rn+n), ierr)
     END DO
  endif
END SUBROUTINE com_halo_l_1_begin

! ========================================================================
subroutine com_halo_1_begin(com, array2D)

  ! General version of the communication routine for 2D nodal fields

  type(com_struct), intent(inout) :: com     
  real(kind=WP), intent(inout) :: array2D(:)

  integer                      :: n, sn, rn, ierr, nreq

  if (npes > 1) then
     sn=com%sPEnum
     rn=com%rPEnum
     nreq = com%nreq

     com%nreq = com%nreq + rn+sn

     ! Expensive, but sorted out after the first timestep
     if (com%nreq > size(com%req)) call enlarge_array(com%nreq, com%req)

     DO n=1,rn         
        call MPI_IRECV(array2D, 1, com%rtype(n,1), com%rPE(n), &
             com%rPE(n), mpi_comm_tsuna, com%req(nreq + n), ierr) 
     END DO
     DO n=1, sn
        call MPI_ISEND(array2D, 1, com%stype(n,1), com%sPE(n), &
             mype, mpi_comm_tsuna, com%req(nreq + rn+n), ierr)
     END DO

  end if

END SUBROUTINE com_halo_1_begin


! ========================================================================
subroutine com_halo_n_begin(com, array2D)

  type(com_struct), intent(inout):: com   
  real(kind=WP), intent(inout) :: array2D(:,:) 
  ! General version of the communication routine for 2D fields
! with several values in each node
 
  integer  :: n, sn, rn
  integer  :: nz, n_val, ierr, nreq

 if (npes > 1) then
    sn=com%sPEnum
    rn=com%rPEnum
    nreq = com%nreq

    com%nreq = com%nreq + rn+sn

    ! Expensive, but sorted out after the first timestep
    if (com%nreq > size(com%req)) call enlarge_array(com%nreq, com%req)
     
    n_val=ubound(array2D,1)

    if (n_val > com%mpitype_maxnval) then
       if (mype==0) then
          print *,'Halo exchange not implemented for',n_val,'values.'
          print *,'You have to add the MPI datatype.'
       endif
       call com_parallel_exit(1)
    endif

    DO n=1,rn    
       call MPI_IRECV(array2D, 1, com%rtype(n,n_val), com%rPE(n), &
            com%rPE(n), mpi_comm_tsuna, com%req(nreq + n), ierr) 
    END DO

    DO n=1, sn
       call MPI_ISEND(array2D, 1, com%stype(n,n_val), com%sPE(n), &
            mype, mpi_comm_tsuna, com%req(nreq + rn+n), ierr)
    END DO

 endif
END SUBROUTINE com_halo_n_begin

! ========================================================================
subroutine com_halo_i_n_begin(com, array2D)

  type(com_struct), intent(inout):: com   
  integer, intent(inout) :: array2D(:,:) 
  ! General version of the communication routine for 2D fields
! with several values in each node
 
  integer  :: n, sn, rn
  integer  :: nz, n_val, ierr, nreq

 if (npes > 1) then

    sn=com%sPEnum
    rn=com%rPEnum
    nreq = com%nreq

    com%nreq = com%nreq + rn+sn

    ! Expensive, but sorted out after the first timestep
    if (com%nreq > size(com%req)) call enlarge_array(com%nreq, com%req)
     
    n_val=ubound(array2D,1)

    if (n_val > com%mpitype_maxnval) then
       if (mype==0) then
          print *,'Halo exchange not implemented for',n_val,'values.'
          print *,'You have to add the MPI datatype.'
       endif
       call com_parallel_exit(1)
    endif

    DO n=1,rn    
       call MPI_IRECV(array2D, 1, com%rtype_i(n,n_val), com%rPE(n), &
            com%rPE(n), mpi_comm_tsuna, com%req(nreq + n), ierr) 
    END DO

    DO n=1, sn
       call MPI_ISEND(array2D, 1, com%stype_i(n,n_val), com%sPE(n), &
            mype, mpi_comm_tsuna, com%req(nreq + rn+n), ierr)
    END DO

 endif
END SUBROUTINE com_halo_i_n_begin

!=======================================
! AND WAITING
!=======================================
 
SUBROUTINE com_halo_end(com)

  type(com_struct), intent(inout)        :: com
  integer                                :: ierr

! Wait for all halo exchanges for the specific halo (nod2D, elem2D, or elem2D_full)
  
  if (npes > 1) then

     ! It should be ok to pass 0 requests, but I could not find a behaviour defined
     ! in the MPI standard. Thus, be on the safe side, check! It should be possible
     ! to issue an additional com_halo_end by accident.
     if (com%nreq>0) call MPI_WAITALL(com%nreq, com%req, MPI_STATUSES_IGNORE, ierr)
     com%nreq = 0
  endif
       
END SUBROUTINE com_halo_end
!
!============================================================================
!
subroutine com_gather_single2single_n(com, arr2D, arr2D_global)

! Make nodal information available to master PE 
!
! Use only with 2D arrays with multiple values 

  type(com_struct), intent(in) :: com 
  real(kind=SP), intent(in)    ::  arr2D(:,:)
  real(kind=SP), intent(out)   ::  arr2D_global(:,:)
  real(kind=SP), allocatable   :: recvbuf(:,:)
  integer        :: req(npes-1)
  integer        :: iA, iZ, num
  INTEGER        :: n_val
  integer        :: n, ierr

  n_val=ubound(arr2D,1)

  IF ( mype == 0 ) THEN
     if (npes>1) then
        allocate(recvbuf(n_val,com%gptr(npes)-1))

        do  n = 1, npes-1
           num   = (com%gptr(n+1) - com%gptr(n))*n_val
           iA = com%gptr(n)
           call MPI_IRECV(recvbuf(1,iA), num, MPI_REAL, n, 2, mpi_comm_tsuna, req(n), ierr)
        enddo
      
        arr2D_global(1:n_val,com%myList(1:com%myDim)) = arr2D(1:n_val,1:com%myDim)
   
        call MPI_WAITALL(npes-1, req, MPI_STATUSES_IGNORE, ierr)
   
        do n = 1, npes - 1
           iA = com%gptr(n)
           iZ = com%gptr(n+1) - 1
           arr2D_global(1:n_val, com%glist(iA : iZ)) = recvbuf(1:n_val, iA : iZ)
        enddo
        
        deallocate(recvbuf)
     else
        arr2D_global(:,:) = arr2D(:,:)
     endif

  ELSE
     call MPI_SEND( arr2D, com%myDim*n_val, MPI_REAL, 0, 2, mpi_comm_tsuna, ierr )
  ENDIF

end subroutine com_gather_single2single_n
!==============================================
subroutine com_gather_single2single_1(com, arr2D, arr2D_global)

! Make nodal information available to master PE 

  type(com_struct), intent(in) :: com 
  real(kind=SP), intent(in)    ::  arr2D(:)
  real(kind=SP), intent(out)   ::  arr2D_global(:)
  real(kind=SP), allocatable   :: recvbuf(:) 
  integer        :: req(npes-1)
  integer        :: iA, iZ, num, n, ierr
  
  IF ( mype == 0 ) THEN
   
     if (npes>1) then
        allocate(recvbuf(com%gptr(npes)-1))
        
        do  n = 1, npes-1
           num   = com%gptr(n+1) - com%gptr(n)
           iA = com%gptr(n)
           call MPI_IRECV(recvbuf(iA), num, MPI_REAL, n, 2, mpi_comm_tsuna, req(n), ierr)
        enddo
        
        arr2D_global(com%myList(1:com%myDim)) = arr2D(1:com%myDim)
        
        call MPI_WAITALL(npes-1, req, MPI_STATUSES_IGNORE, ierr)

        ! For nodes, we could do the copy in one go, as local to global is bijective.
        !    arr2D_global(com%glist(1 : com%gptr(npes)-1)) &
        !                 = recvbuf(1 : com%gptr(npes)-1)
        ! Elements however are kept twice in the halo area, and we have to be careful.
        do n = 1, npes - 1
           iA = com%gptr(n)
           iZ = com%gptr(n+1) - 1
           arr2D_global(com%glist(iA : iZ)) = recvbuf(iA : iZ)
        enddo

        deallocate(recvbuf)
     else
        arr2D_global(:) = arr2D(:)
     endif
  ELSE
     call MPI_SEND( arr2D, com%myDim, MPI_REAL, 0, 2, mpi_comm_tsuna, ierr )
  ENDIF

end subroutine com_gather_single2single_1
!==============================================
!
subroutine com_gather_double2double_n(com, arr2D, arr2D_global)

! Make nodal information available to master PE 
!
! Use only with 2D arrays with multiple values 

  type(com_struct), intent(in) :: com 
  real(kind=DP), intent(in)    ::  arr2D(:,:)
  real(kind=DP), intent(out)   ::  arr2D_global(:,:)
  real(kind=DP), allocatable   :: recvbuf(:,:)
  integer        :: req(npes-1)
  integer        :: iA, iZ, num
  INTEGER        :: n_val
  integer        :: n, ierr

  n_val=ubound(arr2D,1)

  IF ( mype == 0 ) THEN
     if (npes>1) then
        allocate(recvbuf(n_val, com%gptr(npes)-1))

        do  n = 1, npes-1
           num   = (com%gptr(n+1) - com%gptr(n))*n_val
           iA = com%gptr(n)
           call MPI_IRECV(recvbuf(1,iA), num, MPI_DOUBLE_PRECISION, n, 2, mpi_comm_tsuna, req(n), ierr)
        enddo
      
        arr2D_global(1:n_val,com%myList(1:com%myDim)) = arr2D(1:n_val,1:com%myDim)
   
        call MPI_WAITALL(npes-1, req, MPI_STATUSES_IGNORE, ierr)
   
        do n = 1, npes - 1
           iA = com%gptr(n)
           iZ = com%gptr(n+1) - 1
           arr2D_global(1:n_val, com%glist(iA : iZ)) = recvbuf(1:n_val, iA : iZ)
        enddo

        deallocate(recvbuf)
     else
        arr2D_global(:,:) = arr2D(:,:)
     endif

  ELSE
     call MPI_SEND( arr2D, com%myDim*n_val, MPI_DOUBLE_PRECISION, 0, 2, mpi_comm_tsuna, ierr )
  ENDIF

end subroutine com_gather_double2double_n
!==============================================
subroutine com_gather_double2double_1(com, arr2D, arr2D_global)

! Make nodal information available to master PE 

  type(com_struct), intent(in) :: com 
  real(kind=DP), intent(in)    ::  arr2D(:)
  real(kind=DP), intent(out)   ::  arr2D_global(:)
  real(kind=DP), allocatable   :: recvbuf(:)
  integer        :: req(npes-1)
  integer        :: iA, iZ, num, n, ierr
  
  IF ( mype == 0 ) THEN
   
     if (npes>1) then
        allocate(recvbuf(com%gptr(npes)-1))
        
        do  n = 1, npes-1
           num   = com%gptr(n+1) - com%gptr(n)
           iA = com%gptr(n)
           call MPI_IRECV(recvbuf(iA), num, MPI_DOUBLE_PRECISION, n, 2, mpi_comm_tsuna, req(n), ierr)
        enddo
        
        arr2D_global(com%myList(1:com%myDim)) = arr2D(1:com%myDim)
        
        call MPI_WAITALL(npes-1, req, MPI_STATUSES_IGNORE, ierr)
       
        do n = 1, npes - 1
           iA = com%gptr(n)
           iZ = com%gptr(n+1) - 1
           arr2D_global(com%glist(iA : iZ)) = recvbuf(iA : iZ)
        enddo

        deallocate(recvbuf)
     else
        arr2D_global(:) = arr2D(:)
     endif
  ELSE
     call MPI_SEND( arr2D, com%myDim, MPI_DOUBLE_PRECISION, 0, 2, mpi_comm_tsuna, ierr )
  ENDIF

end subroutine com_gather_double2double_1
!==============================================
!
subroutine com_gather_i_n(com, arr2D, arr2D_global)

! Make nodal information available to master PE 
!
! Use only with 2D arrays with multiple values 

  type(com_struct), intent(in) :: com 
  integer, intent(in)    ::  arr2D(:,:)
  integer, intent(out)   ::  arr2D_global(:,:)
  integer, allocatable   :: recvbuf(:,:)
  integer        :: req(npes-1)
  integer        :: iA, iZ, num
  INTEGER        :: n_val
  integer        :: n, ierr
  
  
  n_val=ubound(arr2D,1)

  IF ( mype == 0 ) THEN
     if (npes>1) then
        allocate(recvbuf(n_val,com%gptr(npes)-1))
        
        do  n = 1, npes-1
           num   = (com%gptr(n+1) - com%gptr(n))*n_val
           iA = com%gptr(n)
           call MPI_IRECV(recvbuf(1,iA), num, mpi_integer, n, 2, mpi_comm_tsuna, req(n), ierr)
        enddo
        
        arr2D_global(1:n_val,com%myList(1:com%myDim)) = arr2D(1:n_val,1:com%myDim)
        
        call MPI_WAITALL(npes-1, req, MPI_STATUSES_IGNORE, ierr)
        
        do n = 1, npes - 1
           iA = com%gptr(n)
           iZ = com%gptr(n+1) - 1
           arr2D_global(1:n_val, com%glist(iA : iZ)) = recvbuf(1:n_val, iA : iZ)
        enddo

        deallocate(recvbuf)
     else
        arr2D_global(:,:) = arr2D(:,:)
     endif

  ELSE
     call MPI_SEND( arr2D, com%myDim*n_val, mpi_integer, 0, 2, mpi_comm_tsuna, ierr )
  ENDIF

end subroutine com_gather_i_n
!==============================================
subroutine com_gather_i_1(com, arr2D, arr2D_global)

! Make nodal information available to master PE 

type(com_struct), intent(in) :: com 
integer, intent(in)    ::  arr2D(:)
integer, intent(out)   ::  arr2D_global(:)
integer, allocatable   :: recvbuf(:)
integer                :: req(npes-1)
integer                :: iA, iZ, num, n, ierr

IF ( mype == 0 ) THEN
   
   if (npes>1) then
      allocate(recvbuf(com%gptr(npes)-1))
      
      do  n = 1, npes-1
         num   = com%gptr(n+1) - com%gptr(n)
         iA = com%gptr(n)
         call MPI_IRECV(recvbuf(iA), num, mpi_integer, n, 2, mpi_comm_tsuna, req(n), ierr)
      enddo
      
      arr2D_global(com%myList(1:com%myDim)) = arr2D(1:com%myDim)
   
      call MPI_WAITALL(npes-1, req, MPI_STATUSES_IGNORE, ierr)
   
      do n = 1, npes - 1
         iA = com%gptr(n)
         iZ = com%gptr(n+1) - 1
         arr2D_global(com%glist(iA : iZ)) = recvbuf(iA : iZ)
      enddo

        deallocate(recvbuf)
   else
      arr2D_global(:) = arr2D(:)
   endif

ELSE
   call MPI_SEND( arr2D, com%myDim, mpi_integer, 0, 2, mpi_comm_tsuna, ierr )

ENDIF

end subroutine com_gather_i_1

!=====================================================================

subroutine com_gather_double2single_n(com, arr2D, arr2D_global)

! Make nodal information available to master PE 
!
! Use only with 2D arrays with multiple values 

type(com_struct), intent(in) :: com 
real(kind=dp), intent(in)       :: arr2D(:,:)
real(kind=sp), intent(out)      :: arr2D_global(:,:)
real(kind=sp), allocatable      :: recvbuf(:,:)
integer        :: req(npes-1)
integer        :: iA, iZ, num
INTEGER        :: n_val
integer        :: n, ierr

n_val=ubound(arr2D,1)

IF ( mype == 0 ) THEN
   
   if (npes>1) then
      allocate(recvbuf(n_val,com%gptr(npes)-1))

      do  n = 1, npes-1
         num   = (com%gptr(n+1) - com%gptr(n))*n_val
         iA = com%gptr(n)
         call MPI_IRECV(recvbuf(1,iA), num, MPI_REAL, n, 2, mpi_comm_tsuna, req(n), ierr)
      enddo
      
      arr2D_global(1:n_val,com%myList(1:com%myDim)) = real(arr2D(1:n_val,1:com%myDim),sp)
   
      call MPI_WAITALL(npes-1, req, MPI_STATUSES_IGNORE, ierr)
   
        do n = 1, npes - 1
           iA = com%gptr(n)
           iZ = com%gptr(n+1) - 1
           arr2D_global(1:n_val, com%glist(iA : iZ)) = recvbuf(1:n_val, iA : iZ)
        enddo

        deallocate(recvbuf)
   else
      arr2D_global(:,:) = real(arr2D(:,:),sp)
   endif

ELSE
   call MPI_SEND( real(arr2D,sp), com%myDim*n_val, MPI_REAL, 0, 2, mpi_comm_tsuna, ierr )
ENDIF

end subroutine com_gather_double2single_n
!==============================================
subroutine com_gather_double2single_1(com, arr2D, arr2D_global)

! Make nodal information available to master PE 

type(com_struct), intent(in) :: com 
real(kind=DP), intent(in)       ::  arr2D(:)
real(kind=SP), intent(out)      ::  arr2D_global(:)
real(kind=SP), allocatable      :: recvbuf(:)
integer                         :: req(npes-1)
integer                         :: iA, iZ, num, n, ierr

IF ( mype == 0 ) THEN
   
   if (npes>1) then
      allocate(recvbuf(com%gptr(npes)-1))

      do  n = 1, npes-1
         num   = com%gptr(n+1) - com%gptr(n)
         iA = com%gptr(n)
         call MPI_IRECV(recvbuf(iA), num, MPI_REAL, n, 2, mpi_comm_tsuna, req(n), ierr)
      enddo
      
      arr2D_global(com%myList(1:com%myDim)) = real(arr2D(1:com%myDim),sp)
   
      call MPI_WAITALL(npes-1, req, MPI_STATUSES_IGNORE, ierr)
   
      do n = 1, npes - 1
         iA = com%gptr(n)
         iZ = com%gptr(n+1) - 1
         arr2D_global(com%glist(iA : iZ)) = recvbuf(iA : iZ)
      enddo

        deallocate(recvbuf)
   else
      arr2D_global(:) = real(arr2D(:),sp)
   endif
ELSE
   call MPI_SEND( real(arr2D,sp), com%myDim, MPI_REAL, 0, 2, mpi_comm_tsuna, ierr )
ENDIF

end subroutine com_gather_double2single_1

!=======================================================
! And the other way round: mype==0 distributes global information

subroutine com_scatter_n(com, arr2D_global, arr2D)

  type(com_struct), intent(in) :: com 
  real(kind=WP), intent(in)    :: arr2D_global(:,:)
  real(kind=WP), intent(out)   :: arr2D(:,:)
  real(kind=WP), allocatable   :: sendbuf(:,:)
  integer        :: req(npes-1)
  integer        :: iA, iZ, num
  INTEGER        :: n_val
  integer        :: n, ierr

   n_val=ubound(arr2D,1)

   IF ( mype == 0 ) THEN
      
      if (npes>1) then
         allocate(sendbuf(n_val, com%gmax_len))
         
         do  n = 1, npes-1
            num   = com%gptr(n+1) - com%gptr(n)
            iA = com%gptr(n)
            iZ = com%gptr(n+1) - 1
            
            sendbuf(1:n_val,1:num) = &
                 arr2D_global(1:n_val,com%glist(iA:iZ))
            
            call MPI_SEND(sendbuf, num*n_val, mpi_wp, n, 2, mpi_comm_tsuna, req(n), ierr)
         enddo

         deallocate(sendbuf)
         arr2D(1:n_val,1:com%myDim) = arr2D_global(1:n_val,com%myList(1:com%myDim))
         
!         call MPI_WAITALL(npes-1, req, MPI_STATUSES_IGNORE, ierr)
      else
         arr2D(:,:) = arr2D_global(:,:)
      endif
      
   ELSE
      call MPI_RECV( arr2D, com%myDim*n_val, mpi_wp, 0, 2, mpi_comm_tsuna, MPI_STATUS_IGNORE, ierr )
   ENDIF

end subroutine com_scatter_n

!====================================================

subroutine com_scatter_1(com, arr2D_global, arr2D)

! Make nodal information available to master PE 
!
! Use only with 2D arrays with multiple values 

type(com_struct), intent(in) :: com 
real(kind=WP), intent(in)    :: arr2D_global(:)
real(kind=WP), intent(out)   :: arr2D(:)
real(kind=WP), allocatable   :: sendbuf(:)
integer        :: req(npes-1)
integer        :: iA, iZ, num
integer        :: n, ierr

   IF ( mype == 0 ) THEN
      
      if (npes>1) then
         allocate(sendbuf(com%gmax_len))
         
         do  n = 1, npes-1
            num   = com%gptr(n+1) - com%gptr(n)
            iA = com%gptr(n)
            iZ = com%gptr(n+1) - 1
            
            sendbuf(1:num) = &
                 arr2D_global(com%glist(iA:iZ))
            
            call MPI_SEND(sendbuf, num, mpi_wp, n, 2, mpi_comm_tsuna, req(n), ierr)
         enddo

         deallocate(sendbuf)
         arr2D(1:com%myDim) = arr2D_global(com%myList(1:com%myDim))
         
!         call MPI_WAITALL(npes-1, req, MPI_STATUSES_IGNORE, ierr)
      else
         arr2D(:) = arr2D_global(:)
      endif
      
   ELSE
      call MPI_RECV( arr2D, com%myDim, mpi_wp, 0, 2, mpi_comm_tsuna,  MPI_STATUS_IGNORE, ierr )
   ENDIF

end subroutine com_scatter_1

!=============================================================================


subroutine com_scatter_i_n(com, arr2D_global, arr2D)

  type(com_struct), intent(in) :: com 
  integer, intent(in)          :: arr2D_global(:,:)
  integer, intent(out)         :: arr2D(:,:)
  integer, allocatable         :: sendbuf(:,:)
  integer        :: req(npes-1)
  integer        :: iA, iZ, num
  INTEGER        :: n_val
  integer        :: n, ierr


   n_val=ubound(arr2D,1)

   IF ( mype == 0 ) THEN
      
      if (npes>1) then
         allocate(sendbuf(n_val, com%gmax_len))
         
         do  n = 1, npes-1
            num   = com%gptr(n+1) - com%gptr(n)
            iA = com%gptr(n)
            iZ = com%gptr(n+1) - 1
            
            sendbuf(1:n_val,1:num) = &
                 arr2D_global(1:n_val,com%glist(iA:iZ))
            call MPI_SEND(sendbuf, num*n_val, MPI_INTEGER, n, 2, mpi_comm_tsuna, req(n), ierr)
         enddo

         deallocate(sendbuf)
         arr2D(1:n_val,1:com%myDim) = arr2D_global(1:n_val,com%myList(1:com%myDim))

         !         call MPI_WAITALL(npes-1, req, MPI_STATUSES_IGNORE, ierr)
      else
         arr2D(:,:) = arr2D_global(:,:)
      endif
      
   ELSE
       call MPI_RECV( arr2D, com%myDim*n_val, MPI_INTEGER, 0, 2, mpi_comm_tsuna, MPI_STATUS_IGNORE, ierr )
    ENDIF

end subroutine com_scatter_i_n
!=============================================================
subroutine com_scatter_i_1(com, arr2D_global, arr2D)

! Make nodal information available to master PE 
!
! Use only with 2D arrays with multiple values 

  type(com_struct), intent(in) :: com 
  integer, intent(in)    :: arr2D_global(:)
  integer, intent(out)   :: arr2D(:)
  integer, allocatable   :: sendbuf(:)
  integer        :: req(npes-1)
  integer        :: iA, iZ, num
  integer        :: n, ierr


   IF ( mype == 0 ) THEN
      
      if (npes>1) then
         allocate(sendbuf(com%gmax_len))
         
         do  n = 1, npes-1
            num   = com%gptr(n+1) - com%gptr(n)
            iA = com%gptr(n)
            iZ = com%gptr(n+1) - 1
            
            sendbuf(1:num) = &
                 arr2D_global(com%glist(iA:iZ))
            
            call MPI_SEND(sendbuf, num, MPI_INTEGER, n, 2, mpi_comm_tsuna, req(n), ierr)
         enddo

         deallocate(sendbuf)
         arr2D(1:com%myDim) = arr2D_global(com%myList(1:com%myDim))
         
!         call MPI_WAITALL(npes-1, req, MPI_STATUSES_IGNORE, ierr)
      else
         arr2D(:) = arr2D_global(:)
      endif
      
   ELSE
      call MPI_RECV( arr2D, com%myDim, MPI_INTEGER, 0, 2, mpi_comm_tsuna, MPI_STATUS_IGNORE, ierr )
   ENDIF

end subroutine com_scatter_i_1

!=============================================================

subroutine com_reduce_i_begin(val,com_op)

  integer, intent(inout) :: val(:)
  integer, intent(in)    :: com_op
  integer                :: nval, ierr

  if (npes>1) then
     nval = size(val)
     n_red_req = n_red_req + 1
     if (n_red_req > max_red_req) call enlarge_array(max_red_req+16, reduce_req)
     
     if (mype==0) then
        call MPI_Ireduce(MPI_IN_PLACE, val, nval, MPI_INTEGER, com_op, 0,&
             mpi_comm_tsuna, reduce_req(n_red_req), ierr)
     else
        call MPI_Ireduce(val,          val, nval, MPI_INTEGER, com_op, 0,&
             mpi_comm_tsuna, reduce_req(n_red_req), ierr)
     end if
  endif
end subroutine com_reduce_i_begin

subroutine com_reduce_i_1_begin(val,com_op)

  integer, intent(inout) :: val
  integer, intent(in)    :: com_op
  integer                :: nval, ierr

  if (npes>1) then
     n_red_req = n_red_req + 1
     if (n_red_req > max_red_req) call enlarge_array(max_red_req+16, reduce_req)
     
     if (mype==0) then
        call MPI_Ireduce(MPI_IN_PLACE, val, 1, MPI_INTEGER, com_op, 0,&
             mpi_comm_tsuna, reduce_req(n_red_req), ierr)
     else
        call MPI_Ireduce(val,          val, 1, MPI_INTEGER, com_op, 0,&
             mpi_comm_tsuna, reduce_req(n_red_req), ierr)
     end if
  endif
end subroutine com_reduce_i_1_begin

subroutine com_reduce_sp_begin(val,com_op)

  real(kind=sp), intent(inout) :: val(:)
  integer, intent(in)          :: com_op
  integer                      :: nval, ierr
  if (npes > 1) then
     nval = size(val)
     n_red_req = n_red_req + 1
     if (n_red_req > max_red_req) call enlarge_array(max_red_req+16, reduce_req)
     if (mype==0) then
        call MPI_Ireduce(MPI_IN_PLACE, val, nval, MPI_REAL, com_op, 0,&
             mpi_comm_tsuna, reduce_req(n_red_req), ierr)
     else
        call MPI_Ireduce(val,          val, nval, MPI_REAL, com_op, 0,&
             mpi_comm_tsuna, reduce_req(n_red_req), ierr)
     end if
  end if
end subroutine com_reduce_sp_begin

subroutine com_reduce_dp_begin(val,com_op)

  real(kind=dp), intent(inout) :: val(:)
  integer, intent(in)          :: com_op
  integer                      :: ierr, nval

  if (npes > 1)  then
     nval = size(val)
     n_red_req = n_red_req + 1
     if (n_red_req > max_red_req) call enlarge_array(max_red_req+16, reduce_req)
     
     if (mype==0) then
        call MPI_Ireduce(MPI_IN_PLACE, val, nval, MPI_DOUBLE_PRECISION, com_op, 0,&
             mpi_comm_tsuna, reduce_req(n_red_req), ierr)
     else
        call MPI_Ireduce(val,          val, nval, MPI_DOUBLE_PRECISION, com_op, 0,&
             mpi_comm_tsuna, reduce_req(n_red_req), ierr)
     endif
  end if
end subroutine com_reduce_dp_begin

subroutine com_reduce_sp_1_begin(val,com_op)

  real(kind=sp), intent(inout) :: val
  integer, intent(in)          :: com_op
  integer                      :: ierr
  if (npes > 1) then
     n_red_req = n_red_req + 1
     if (n_red_req > max_red_req) call enlarge_array(max_red_req+16, reduce_req)
     if (mype==0) then
        call MPI_Ireduce(MPI_IN_PLACE, val, 1, MPI_REAL, com_op, 0,&
             mpi_comm_tsuna, reduce_req(n_red_req), ierr)
     else
        call MPI_Ireduce(val,          val, 1, MPI_REAL, com_op, 0,&
             mpi_comm_tsuna, reduce_req(n_red_req), ierr)
     endif
  end if
end subroutine com_reduce_sp_1_begin

subroutine com_reduce_dp_1_begin(val,com_op)

  real(kind=dp), intent(inout) :: val
  integer, intent(in)          :: com_op
  integer                      :: nval, ierr

  if (npes > 1)  then
     n_red_req = n_red_req + 1
     if (n_red_req > max_red_req) call enlarge_array(max_red_req+16, reduce_req)

     if (mype == 0) then
        call MPI_Ireduce(MPI_IN_PLACE, val, 1, MPI_DOUBLE_PRECISION, com_op, 0,&
             mpi_comm_tsuna, reduce_req(n_red_req), ierr)
     else
        call MPI_Ireduce(val,          val, 1, MPI_DOUBLE_PRECISION, com_op, 0,&
             mpi_comm_tsuna, reduce_req(n_red_req), ierr)
     end if
  end if
end subroutine com_reduce_dp_1_begin

!==================================================================


!=============================================================

subroutine com_allreduce_i_begin(val,com_op)

  integer, intent(inout) :: val(:)
  integer, intent(in)    :: com_op
  integer                :: nval, ierr

  if (npes>1) then
     nval = size(val)
     n_red_req = n_red_req + 1
     if (n_red_req > max_red_req) call enlarge_array(max_red_req+16, reduce_req)
     
     call MPI_Iallreduce(MPI_IN_PLACE, val, nval, MPI_INTEGER, com_op,&
          mpi_comm_tsuna, reduce_req(n_red_req), ierr)
  endif
end subroutine com_allreduce_i_begin

subroutine com_allreduce_i_1_begin(val,com_op)

  integer, intent(inout) :: val
  integer, intent(in)    :: com_op
  integer                :: nval, ierr

  if (npes>1) then
     n_red_req = n_red_req + 1
     if (n_red_req > max_red_req) call enlarge_array(max_red_req+16, reduce_req)
     call MPI_Iallreduce(MPI_IN_PLACE, val, 1, MPI_INTEGER, com_op,&
          mpi_comm_tsuna, reduce_req(n_red_req), ierr)

  endif
end subroutine com_allreduce_i_1_begin

subroutine com_allreduce_sp_begin(val,com_op)

  real(kind=sp), intent(inout) :: val(:)
  integer, intent(in)          :: com_op
  integer                      :: nval, ierr
  if (npes > 1) then
     nval = size(val)
     n_red_req = n_red_req + 1
     if (n_red_req > max_red_req) call enlarge_array(max_red_req+16, reduce_req)
     
     call MPI_Iallreduce(MPI_IN_PLACE, val, nval, MPI_REAL, com_op,&
          mpi_comm_tsuna, reduce_req(n_red_req), ierr)
  end if
end subroutine com_allreduce_sp_begin

subroutine com_allreduce_dp_begin(val,com_op)

  real(kind=dp), intent(inout) :: val(:)
  integer, intent(in)          :: com_op
  integer                      :: ierr, nval

  if (npes > 1)  then
     nval = size(val)
     n_red_req = n_red_req + 1
     if (n_red_req > max_red_req) call enlarge_array(max_red_req+16, reduce_req)
     
     call MPI_Iallreduce(MPI_IN_PLACE, val, nval, MPI_DOUBLE_PRECISION, com_op,&
          mpi_comm_tsuna, reduce_req(n_red_req), ierr)
  end if
end subroutine com_allreduce_dp_begin

subroutine com_allreduce_sp_1_begin(val,com_op)

  real(kind=sp), intent(inout) :: val
  integer, intent(in)          :: com_op
  integer                      :: ierr
  if (npes > 1) then
     n_red_req = n_red_req + 1
     if (n_red_req > max_red_req) call enlarge_array(max_red_req+16, reduce_req)
     
     call MPI_Iallreduce(MPI_IN_PLACE, val, 1, MPI_REAL, com_op,&
          mpi_comm_tsuna, reduce_req(n_red_req), ierr)
  end if
end subroutine com_allreduce_sp_1_begin

subroutine com_allreduce_dp_1_begin(val,com_op)

  real(kind=dp), intent(inout) :: val
  integer, intent(in)          :: com_op
  integer                      :: nval, ierr

  if (npes > 1)  then
     n_red_req = n_red_req + 1
     if (n_red_req > max_red_req) call enlarge_array(max_red_req+16, reduce_req)
     call MPI_Iallreduce(MPI_IN_PLACE, val, 1, MPI_DOUBLE_PRECISION, com_op,&
          mpi_comm_tsuna, reduce_req(n_red_req), ierr)
  end if
end subroutine com_allreduce_dp_1_begin

!==============================================================

subroutine com_reduce_end
  integer :: ierr
  
  if (npes > 0) then
      if (n_red_req>0) call MPI_Waitall(n_red_req, reduce_req, MPI_STATUSES_IGNORE, ierr)
     n_red_req=0
  endif
  
end subroutine com_reduce_end

!=============================================================

subroutine enlarge_array(new_dim, array)

  integer, intent(in)                 :: new_dim
  integer, allocatable, intent(inout) :: array(:)
  
  integer, allocatable                :: aux(:)
  integer                             :: n_old

  ! This is an expensive routine, however, in the course of the first
  ! time step, the arrays reach their final size.
  
  n_old = size(array)
  allocate(aux(n_old))
  aux(1:n_old) = array(1:n_old)
  deallocate(array)

  allocate(array(new_dim))
  array(1:n_old) = aux(1:n_old)
 
end subroutine enlarge_array


!======================================================
#else
! Dummy routines for no MPI

!======================================================
subroutine com_halo_i_1_begin(com, array2D)
  integer, intent(in) :: com     
  integer, intent(in) :: array2D(:)
END SUBROUTINE com_halo_i_1_begin

subroutine com_halo_i_n_begin(com, array2D)
  integer, intent(in) :: com     
  integer, intent(in) :: array2D(:,:)
END SUBROUTINE com_halo_i_n_begin

subroutine com_halo_1_begin(com, array2D)
  integer, intent(in)       :: com     
  real(kind=WP), intent(in) :: array2D(:)
END SUBROUTINE com_halo_1_begin

subroutine com_halo_n_begin(com, array2D)
  integer, intent(in)       :: com   
  real(kind=WP), intent(in) :: array2D(:,:)
END SUBROUTINE com_halo_n_begin

subroutine com_halo_l_1_begin(com, array2D)
  integer, intent(in)       :: com   
  logical, intent(in)       :: array2D(:)
END SUBROUTINE com_halo_l_1_begin
 
SUBROUTINE com_halo_end(com)
  integer, intent(in) :: com
END SUBROUTINE com_halo_end

subroutine com_gather_double2double_n(com, arr2D, arr2D_global)
  integer, intent(in)        :: com 
  real(kind=DP), intent(in)  ::  arr2D(:,:)
  real(kind=DP), intent(out) ::  arr2D_global(:,:)
  arr2D_global(:,:) = arr2D(:,:)
end subroutine com_gather_double2double_n

subroutine com_gather_double2double_1(com, arr2D, arr2D_global)
  integer, intent(in)        :: com 
  real(kind=DP), intent(in)  ::  arr2D(:)
  real(kind=DP), intent(out) ::  arr2D_global(:)
  arr2D_global(:) = arr2D(:)
end subroutine com_gather_double2double_1

subroutine com_gather_single2single_n(com, arr2D, arr2D_global)
  integer, intent(in)        :: com 
  real(kind=SP), intent(in)  ::  arr2D(:,:)
  real(kind=SP), intent(out) ::  arr2D_global(:,:)
  arr2D_global(:,:) = arr2D(:,:)
end subroutine com_gather_single2single_n

subroutine com_gather_single2single_1(com, arr2D, arr2D_global)
  integer, intent(in)        :: com 
  real(kind=SP), intent(in)  ::  arr2D(:)
  real(kind=SP), intent(out) ::  arr2D_global(:)
  arr2D_global(:) = arr2D(:)
end subroutine com_gather_single2single_1

subroutine com_gather_i_n(com, arr2D, arr2D_global)
  integer, intent(in)  :: com 
  integer, intent(in)  ::  arr2D(:,:)
  integer, intent(out) ::  arr2D_global(:,:)
  arr2D_global(:,:) = arr2D(:,:)
end subroutine com_gather_i_n

subroutine com_gather_i_1(com, arr2D, arr2D_global)
  integer, intent(in)  :: com 
  integer, intent(in)  :: arr2D(:)
  integer, intent(out) :: arr2D_global(:)
  arr2D_global(:) = arr2D(:)
end subroutine com_gather_i_1

subroutine com_gather_double2single_n(com, arr2D, arr2D_global)
  integer, intent(in)        :: com 
  real(kind=dp), intent(in)  :: arr2D(:,:)
  real(kind=sp), intent(out) :: arr2D_global(:,:)
  arr2D_global(:,:) = real(arr2D(:,:),sp)
end subroutine com_gather_double2single_n

subroutine com_gather_double2single_1(com, arr2D, arr2D_global)
  integer, intent(in)        :: com 
  real(kind=DP), intent(in)  :: arr2D(:)
  real(kind=SP), intent(out) :: arr2D_global(:)
  arr2D_global(:) = real(arr2D(:),sp)
end subroutine com_gather_double2single_1


subroutine com_scatter_n(com, arr2D_global, arr2D)
  integer, intent(in) :: com 
  real(kind=WP), intent(in)    :: arr2D_global(:,:)
  real(kind=WP), intent(in)   :: arr2D(:,:)
end subroutine com_scatter_n

subroutine com_scatter_1(com, arr2D_global, arr2D)
  integer, intent(in) :: com 
  real(kind=WP), intent(in)    :: arr2D_global(:)
  real(kind=WP), intent(in)   :: arr2D(:)
end subroutine com_scatter_1

subroutine com_scatter_i_n(com, arr2D_global, arr2D)
  integer, intent(in) :: com 
  integer, intent(in)          :: arr2D_global(:,:)
  integer, intent(in)         :: arr2D(:,:)
end subroutine com_scatter_i_n

subroutine com_scatter_i_1(com, arr2D_global, arr2D)
  integer, intent(in) :: com 
  integer, intent(in)    :: arr2D_global(:)
  integer, intent(in)   :: arr2D(:)
end subroutine com_scatter_i_1

subroutine com_reduce_i_begin(val,com_op)
  integer, intent(in) :: val(:)
  integer, intent(in) :: com_op
end subroutine com_reduce_i_begin

subroutine com_reduce_sp_begin(val,com_op)
  real(kind=sp), intent(in) :: val(:)
  integer, intent(in)       :: com_op
end subroutine com_reduce_sp_begin

subroutine com_reduce_dp_begin(val,com_op)
  real(kind=dp), intent(in) :: val(:)
  integer, intent(in)       :: com_op
end subroutine com_reduce_dp_begin

subroutine com_reduce_i_1_begin(val,com_op)
  integer, intent(in) :: val
  integer, intent(in) :: com_op
end subroutine com_reduce_i_1_begin

subroutine com_reduce_sp_1_begin(val,com_op)
  real(kind=sp), intent(in) :: val
  integer, intent(in)       :: com_op
end subroutine com_reduce_sp_1_begin

subroutine com_reduce_dp_1_begin(val,com_op)
  real(kind=dp), intent(in) :: val
  integer, intent(in)       :: com_op
end subroutine com_reduce_dp_1_begin

!=============================================================

subroutine com_allreduce_i_begin(val,com_op)
  integer, intent(inout) :: val(:)
  integer, intent(in)    :: com_op
end subroutine com_allreduce_i_begin

subroutine com_allreduce_i_1_begin(val,com_op)
  integer, intent(inout) :: val
  integer, intent(in)    :: com_op
end subroutine com_allreduce_i_1_begin

subroutine com_allreduce_sp_begin(val,com_op)
  real(kind=sp), intent(inout) :: val(:)
  integer, intent(in)          :: com_op
end subroutine com_allreduce_sp_begin

subroutine com_allreduce_dp_begin(val,com_op)
  real(kind=dp), intent(inout) :: val(:)
  integer, intent(in)          :: com_op
end subroutine com_allreduce_dp_begin

subroutine com_allreduce_sp_1_begin(val,com_op)
  real(kind=sp), intent(inout) :: val
  integer, intent(in)          :: com_op
end subroutine com_allreduce_sp_1_begin

subroutine com_allreduce_dp_1_begin(val,com_op)
  real(kind=dp), intent(inout) :: val
  integer, intent(in)          :: com_op
end subroutine com_allreduce_dp_1_begin

!==============================================================

subroutine com_reduce_end
end subroutine com_reduce_end



#endif

end module mpi_communication
