module mpi_parallel

  use data_types
  implicit none

  private
  public :: myDim_nod2D,  eDim_nod2D
  public :: myDim_elem2D, eDim_elem2D, exDim_elem2D
  public :: myDim_edg2D,  eDim_edg2D
  public :: com_nod2D,     com_elem2D, com_elem2D_full

  public :: mype, npes

  public :: com_parallel_init, com_parallel_exit
  public :: walltime
  
#ifdef USE_MPI

  public :: com_struct, mpi_comm_tsuna
  public :: mpi_wp, com_printstruct, n_red_req, reduce_req, max_red_req
  public :: com_read_elements_on_subdomains, com_read_subdomains
  
  include "mpif.h"
  
  ! As TsunAWI can be compiled with double or single precision
  ! (or whatever), we have to specify the default MPI datatype.
  ! If we had different precision for different arrays, consider to make
  ! it a component of the com_struct type!
  integer :: mpi_wp  ! Usually, MPI_FLOAT or MPI_DOUBLE_PRECISION
  
  integer :: mpi_comm_tsuna

  ! Each PE has local-to-global index mapping
  integer, allocatable :: myList_nod2D(:)
  integer, allocatable :: myList_elem2D(:)
  
  ! Contains the information for halo exchange with neighbor PEs
  type com_struct
     ! general
     character(len=12)      :: name='default'  ! Handy for debugging: 
                                               ! "elem_2D", "nod2D", "elem2D_full"

     integer                :: mpitype_maxnval=-1 ! number of values per node/element
                                                  ! for which an mpi_datatype is set up
     ! receive part
     integer                :: rPEnum       ! the number of PE I receive info from 
     integer, allocatable   :: rPE(:)       ! their list
     integer, allocatable   :: rptr(:)      ! pointer to the list of nodes/elems, rptr(rPEnum+1)
     integer, allocatable   :: rlist(:)     ! the list of nodes/elems
     integer, allocatable   :: rtype(:,:)   ! mpitypes for exchange of 2D field
                                            ! in working precision (for 1 to maxval values)
                                            ! rtype(rPEnum, mpitype_maxval)
     
     integer, allocatable   :: rtype_i(:,:) ! mpitype for exchange of 2D integer field
     integer, allocatable   :: rtype_l(:)   ! mpitype for exchange of 2D logical field
                                           
     ! send part
     integer                :: sPEnum  
     integer, allocatable   :: sPE(:)
     integer, allocatable   :: sptr(:)
     integer, allocatable   :: slist(:)
     integer, allocatable   :: stype(:,:)                                           
     integer, allocatable   :: stype_i(:,:)                                           
     integer, allocatable   :: stype_l(:)
     
     ! wait part
     integer, allocatable   :: req(:)    ! requests for MPI_Wait
     integer                :: nreq      ! number of requests for MPI_Wait
                                         ! (to combine halo exchange of several fields)

     ! Now, information for the global mesh
     ! These are only set up on the PE(s) doing the IO
     
     integer              :: myDim     ! myDim_nod2D or myDim_elem2D
     integer              :: eDim      !  eDim_nod2D or  eDim_elem2D
     integer              :: exDim     !           0 or exDim_elem2D
     integer              :: globalDim ! nod2D  or elem2D
     integer, allocatable :: mylist(:) ! myList_nod2D or myList_elem2D
     integer              :: gmax_len  ! rember maximum number to be send, 
                                       ! to easier allocate buffer
     integer, allocatable :: gptr(:)   ! for each PE, index in list
     integer, allocatable :: glist(:)  ! local-to-global for each PE consecutively
  end type com_struct

  type(com_struct) :: com_nod2D
  type(com_struct) :: com_elem2D
  type(com_struct) :: com_elem2D_full  

  ! We also need non-blocking reduce operations. To handle the wait inside
  ! the wrapper, we need the requests.
  integer, allocatable :: reduce_req(:)
  integer              :: n_red_req, max_red_req

#else
  ! dummy definition to avoid to many ifdefs in the main code
  integer :: com_nod2D, com_elem2D, com_elem2D_full
  
#endif

  ! General part, filled with dummy values for non-MPI code,
  integer :: mype = 0
  integer :: npes = 1
  integer :: myDim_nod2D,  eDim_nod2D  = 0
  integer :: myDim_elem2D, eDim_elem2D = 0, exDim_elem2D = 0
  integer :: myDim_edg2D,  eDim_edg2D  = 0

  
contains

  ! Dummy subroutines for no MPI follow below
#ifdef USE_MPI

  ! In case we do not have OpenMP and omp_get_wtime()
  real(kind=dp) function walltime()
    walltime = MPI_Wtime() 
  end function walltime    

  subroutine com_parallel_init  
!$    USE OMP_LIB
    
    integer :: ierr
    integer :: mpithread_provided, mpithread_required
    
    mpi_comm_tsuna = MPI_COMM_WORLD

    mpithread_required = MPI_THREAD_SINGLE
!$OMP PARALLEL    
!$ if (omp_get_num_threads()>1 )  mpithread_required = MPI_THREAD_FUNNELED
!$OMP END PARALLEL
   ! At this point, we cannot determine the number of MPI tasks:
   ! npes=1 works with any number of OpenMP threads and MPI_THREAD_SINGLE would 
   ! be sufficient.

    call MPI_Init_Thread(MPI_THREAD_MULTIPLE, mpithread_provided, ierr)
    call MPI_Comm_Size(mpi_comm_tsuna, npes, ierr)
    call MPI_Comm_Rank(mpi_comm_tsuna, mype, ierr)

!$ if (mpithread_required == MPI_THREAD_FUNNELED .and. &
!$     mpithread_provided == MPI_THREAD_SINGLE   .and. &
!$    npes > 1  ) then
!$    if (mype==0) then
!$      print *,'MPI installation only supports MPI_THREAD_SINGLE.'
!$      print *,'For hybrid runs with multiple MPI tasks _and_ multiple'
!$      print *,'OpenMP threads, at least MPI_THREAD_FUNNELED is required.'
!$      print *,'Please run with OMP_NUM_THREADS=1 or recompile w/o MPI.'
!$    endif
!$    call com_parallel_exit(1)
!$ endif
    
    select case (wp)
    case (dp)
       mpi_wp = MPI_DOUBLE_PRECISION
    case (sp)
       mpi_wp = MPI_REAL
    case default
       if (mype==0) &
            print *,'No MPI communications implemented for floating point precision wp=',wp 
       call com_parallel_exit(1)
    end select

    ! Initialize request array for non-blocking reduce operations
    max_red_req = 16
    allocate(reduce_req(max_red_req))  ! can be enlarged if needed
    n_red_req = 0
    
  end subroutine com_parallel_init


  subroutine com_setup_haloexchange(com)

    implicit none

    type(com_struct), intent(inout) :: com

  integer   n, offset
  integer :: i, max_nb, nb, nini, nend, nval, ierr
  integer, allocatable :: blocklen(:),     displace(:)
  integer, allocatable :: blocklen_tmp(:), displace_tmp(:)

  !
  ! In the distributed memory version, most of the job is already done 
  ! at the initialization phase and is taken into account in read_mesh
  ! routine. Here, MPI datatypes are built and buffers for MPI wait requests
  ! are allocated. 

   if (npes > 1) then
            
!================================================
      ! MPI REQUEST BUFFERS 
      ! Initial allocation, it can be enlarged later.
      ! 3* full send/recv should be sufficient for the start. 
!================================================
      allocate(com%req(3*com%rPEnum + 3*com%sPEnum))


      ! If no value is set, We assume up to 8 values in each node/element of the field
      if (com%mpitype_maxnval <= 0) com%mpitype_maxnval = 8
      
   ! Build MPI Data types for halo exchange: 

      allocate(com%rtype(  com%rPEnum, com%mpitype_maxnval))     ! 2D
      allocate(com%stype(  com%sPEnum, com%mpitype_maxnval))
      allocate(com%rtype_i(com%rPEnum, com%mpitype_maxnval))   ! 2D integer
      allocate(com%stype_i(com%sPEnum, com%mpitype_maxnval))   
      allocate(com%rtype_l(com%rPEnum))                        ! 2D logical
      allocate(com%stype_l(com%sPEnum))   

  
      ! Upper limit for the length of the local interface between the neighbor PEs 
      max_nb = max(maxval(com%rptr(2:com%rPEnum+1) - com%rptr(1:com%rPEnum)), &
                   maxval(com%sptr(2:com%sPEnum+1) - com%sptr(1:com%sPEnum)))

      allocate(displace(max_nb),     blocklen(max_nb))
      allocate(displace_tmp(max_nb), blocklen_tmp(max_nb))

      do n=1,com%rPEnum
         nb = 1
         nini = com%rptr(n)
         nend = com%rptr(n+1) - 1
         displace(:) = 0
         displace(1) = com%rlist(nini) -1  ! C counting, start at 0
         blocklen(:) = 1
         do i=nini+1, nend
            if (com%rlist(i) /= com%rlist(i-1) + 1) then  
               ! New block
               nb = nb+1
               displace(nb) = com%rlist(i) -1
            else
               blocklen(nb) = blocklen(nb)+1
            endif
         enddo


         DO nval = 1,com%mpitype_maxnval

            blocklen_tmp(1:nb) = blocklen(1:nb)*nval
            displace_tmp(1:nb) = displace(1:nb)*nval

            call MPI_TYPE_INDEXED(nb, blocklen_tmp, displace_tmp, mpi_wp, & 
                                 com%rtype(n,nval),  ierr)
            call MPI_TYPE_COMMIT(com%rtype(n,nval),  ierr)
            
            call MPI_TYPE_INDEXED(nb, blocklen_tmp, displace_tmp, MPI_INTEGER, & 
                                 com%rtype_i(n,nval),  ierr)
            call MPI_TYPE_COMMIT(com%rtype_i(n,nval),  ierr)

            
            if (nval==1) then
               call MPI_TYPE_INDEXED(nb, blocklen_tmp, displace_tmp, MPI_LOGICAL, & 
                                    com%rtype_l(n),  ierr)
               call MPI_TYPE_COMMIT(com%rtype_l(n),  ierr)
            endif
         ENDDO
      ENDDO
   

      do n=1,com%sPEnum
         nb = 1
         nini = com%sptr(n)
         nend = com%sptr(n+1) - 1
         displace(:) = 0
         displace(1) = com%slist(nini) -1  ! C counting, start at 0
         blocklen(:) = 1
         do i=nini+1, nend
            if (com%slist(i) /= com%slist(i-1) + 1) then  
               ! New block
               nb = nb+1
               displace(nb) = com%slist(i) -1
            else
               blocklen(nb) = blocklen(nb)+1
            endif
         enddo


         DO nval=1,com%mpitype_maxnval

            blocklen_tmp(1:nb) = blocklen(1:nb)*nval 
            displace_tmp(1:nb) = displace(1:nb)*nval

            call MPI_TYPE_INDEXED(nb, blocklen_tmp, displace_tmp, mpi_wp, & 
                                 com%stype(n,nval),  ierr)
            call MPI_TYPE_COMMIT(com%stype(n,nval),  ierr)
            
            call MPI_TYPE_INDEXED(nb, blocklen_tmp, displace_tmp, MPI_INTEGER, & 
                                 com%stype_i(n,nval),  ierr)
            call MPI_TYPE_COMMIT(com%stype_i(n,nval),  ierr)

            if (nval==1) then
               call MPI_TYPE_INDEXED(nb, blocklen_tmp, displace_tmp, MPI_LOGICAL, & 
                                    com%stype_l(n),  ierr)
               call MPI_TYPE_COMMIT(com%stype_l(n),  ierr)
            endif
         ENDDO
         
      enddo

      deallocate(blocklen,     displace)
      deallocate(blocklen_tmp, displace_tmp)

   endif

end subroutine com_setup_haloexchange

subroutine com_setup_scatterGatherLists

  implicit none
  
  integer :: n2D, e2D, sum_loc_elem2D
  integer :: n, estart, nstart, ierr

  if (mype==0) then

     if (npes > 1) then

        com_nod2D%gmax_len  = 0
        com_elem2D%gmax_len = 0
        
        allocate(com_nod2D%gptr(npes))
        allocate(com_elem2D%gptr(npes))

        com_nod2D%gptr(1)  = 1
        com_elem2D%gptr(1) = 1
        
        do n=1, npes-1
           call MPI_RECV(n2D, 1, MPI_INTEGER, n, 0, mpi_comm_tsuna, MPI_STATUS_IGNORE, ierr )
           call MPI_RECV(e2D, 1, MPI_INTEGER, n, 1, mpi_comm_tsuna, MPI_STATUS_IGNORE, ierr )

           com_nod2D%gptr(n+1)  = com_nod2D%gptr(n)  + n2D
           com_elem2D%gptr(n+1) = com_elem2D%gptr(n) + e2D

           ! Helpful to later allocate send/recv buffer of sufficient size
           com_nod2D%gmax_len  = max(com_nod2D%gmax_len,  n2D)
           com_elem2D%gmax_len = max(com_elem2D%gmax_len, e2D)
        enddo



        allocate(com_nod2D%glist(com_nod2D%gptr(npes)))   ! this should be nod2D - myDim_nod2D
        allocate(com_elem2D%glist(com_elem2D%gptr(npes))) ! this is > elem2D, because the elements overlap.
                                                      ! Consider optimization: avoid multiple communication
                                                      ! of the same elem from different PEs.

        do n=1, npes-1
           nstart = com_nod2D%gptr(n)
           n2D    = com_nod2D%gptr(n+1) - com_nod2D%gptr(n) 
           call MPI_RECV(com_nod2D%glist(nstart), n2D, MPI_INTEGER, n, 2, mpi_comm_tsuna, &
                MPI_STATUS_IGNORE, ierr )
           
           estart = com_elem2D%gptr(n)
           e2D    = com_elem2D%gptr(n+1) - com_elem2D%gptr(n)
           call MPI_RECV(com_elem2D%glist(estart),e2D, MPI_INTEGER, n, 3, mpi_comm_tsuna, &
                MPI_STATUS_IGNORE, ierr ) 
        enddo
     end if
  else

     call MPI_SEND(myDim_nod2D,       1,            MPI_INTEGER, 0, 0, mpi_comm_tsuna, ierr )
     call MPI_SEND(myDim_elem2D,      1,            MPI_INTEGER, 0, 1, mpi_comm_tsuna, ierr )
     call MPI_SEND(com_nod2D%myList,  myDim_nod2D,  MPI_INTEGER, 0, 2, mpi_comm_tsuna, ierr )
     call MPI_SEND(com_elem2D%myList, myDim_elem2D, MPI_INTEGER, 0, 3, mpi_comm_tsuna, ierr )
  endif
  
end subroutine com_setup_scatterGatherLists

subroutine com_parallel_exit(ierr)

  integer, intent(in) :: ierr
  integer             :: mpierr
  
  call MPI_Finalize(mpierr)

  !NR: Problem: "stop ierr" with variable integer ierr is Fortran 2018,
  !NR: but not supported in ifort as of 19.1.2. Workaround:  

  if (mpierr == 0 .and. ierr == 0) then
     stop 0
  else
     if (  ierr /= 0) print *,'TsunAWI failed with error code',ierr 
     if (mpierr /= 0) print *,'MPI_Finalize failed with error code',mpierr
     stop 1     
  end if
  
end subroutine com_parallel_exit

!=================================================

!===============================================================

subroutine com_read_elements_on_subdomains(MeshPath, elem_nod)

  character(len=200), intent(in) :: MeshPath
  integer, intent(out)           :: elem_nod(3, myDim_elem2D+eDim_elem2D+eXDim_elem2D)

  character(len=10)              :: mype_string,npes_string
  character(len=300)             :: DistMeshPath
  integer                        :: infile_elem, el

  write(mype_string,'(i5.5)') mype  
  write(npes_string,'(I10)') npes
  DistMeshPath = trim(MeshPath)//'dist_'//trim(ADJUSTL(npes_string))//'/'
  open(newunit=infile_elem, file=trim(DistMeshPath)//'elem2d_'//trim(mype_string)//'.out')

  do el = 1, myDim_elem2D+eDim_elem2D+eXDim_elem2D
     read(infile_elem, *) elem_nod(1:3,el)
  end do
  close(infile_elem)
  
  
end subroutine com_read_elements_on_subdomains
  
  
!===============================================================

subroutine com_read_subdomains(MeshPath)

  character(len=200), intent(in) :: MeshPath
  INTEGER                        :: infile_cominfo, infile_mylist
  character(len=10)              :: mype_string,npes_string
  character(len=300)             :: DistMeshPath
  integer                        :: n

  com_nod2D%name       = "nod2D"
  com_elem2D%name      = "elem2D"
  com_elem2D_full%name = "elem2D_full"
  
  write(mype_string,'(i5.5)') mype  
  write(npes_string,'(I10)') npes
  DistMeshPath = trim(MeshPath)//'dist_'//trim(ADJUSTL(npes_string))//'/'
  
 !===========================
  ! Lists of nodes and elements in global indexing. 
  ! every proc reads its file
  !===========================
  
  open(newunit=infile_mylist, file=trim(DistMeshPath)//'my_list'//trim(mype_string)//'.out')
  read(infile_mylist,*) n

  read(infile_mylist,*) myDim_nod2D
  read(infile_mylist,*) eDim_nod2D
  allocate(com_nod2D%myList(myDim_nod2D+eDim_nod2D)) 	 
  read(infile_mylist,*) com_nod2D%myList

  read(infile_mylist,*) myDim_elem2D
  read(infile_mylist,*) eDim_elem2D
  read(infile_mylist,*) eXDim_elem2D
  allocate(com_elem2D%myList(myDim_elem2D+eDim_elem2D+eXDim_elem2D))
  read(infile_mylist,*) com_elem2D%myList
  
  close(infile_mylist)
  
  if (mype==0) print *,'myLists are read'
  
  ! ==============================
  ! Communication information
  ! every proc reads its file
  ! ==============================

  open(newunit=infile_cominfo, file=trim(DistMeshPath)//'com_info'//trim(mype_string)//'.out')
  read(infile_cominfo,*) n
  read(infile_cominfo,*) com_nod2D%rPEnum
  allocate(com_nod2D%rPE(com_nod2D%rPEnum))
  allocate(com_nod2D%rptr(com_nod2D%rPEnum+1))
  read(infile_cominfo,*) com_nod2D%rPE(1:com_nod2D%rPEnum)
  read(infile_cominfo,*) com_nod2D%rptr(1:com_nod2D%rPEnum+1)
  ALLOCATE(com_nod2D%rlist(eDim_nod2D))
  read(infile_cominfo,*) com_nod2D%rlist

  read(infile_cominfo,*) com_nod2D%sPEnum
  allocate(com_nod2D%sPE(com_nod2D%sPEnum))
  allocate(com_nod2D%sptr(com_nod2D%sPEnum+1))
  read(infile_cominfo,*) com_nod2D%sPE(1:com_nod2D%sPEnum)
  read(infile_cominfo,*) com_nod2D%sptr(1:com_nod2D%sPEnum+1)
  n=com_nod2D%sptr(com_nod2D%sPEnum+1)-1
  ALLOCATE(com_nod2D%slist(n))
  read(infile_cominfo,*) com_nod2D%slist

  read(infile_cominfo,*) com_elem2D%rPEnum
  allocate(com_elem2D%rPE(com_elem2D%rPEnum))
  allocate(com_elem2D%rptr(com_elem2D%rPEnum+1))
  read(infile_cominfo,*) com_elem2D%rPE(1:com_elem2D%rPEnum)
  read(infile_cominfo,*) com_elem2D%rptr(1:com_elem2D%rPEnum+1)
  ALLOCATE(com_elem2D%rlist(eDim_elem2D))
  read(infile_cominfo,*) com_elem2D%rlist

  read(infile_cominfo,*) com_elem2D%sPEnum
  allocate(com_elem2D%sPE(com_elem2D%sPEnum))
  allocate(com_elem2D%sptr(com_elem2D%sPEnum+1))
  read(infile_cominfo,*) com_elem2D%sPE(1:com_elem2D%sPEnum)
  read(infile_cominfo,*) com_elem2D%sptr(1:com_elem2D%sPEnum+1)
  n=com_elem2D%sptr(com_elem2D%sPEnum+1)-1
  ALLOCATE(com_elem2D%slist(n))
  read(infile_cominfo,*) com_elem2D%slist

  read(infile_cominfo,*) com_elem2D_full%rPEnum
  allocate(com_elem2D_full%rPE(com_elem2D_full%rPEnum))
  allocate(com_elem2D_full%rptr(com_elem2D_full%rPEnum+1))
  read(infile_cominfo,*) com_elem2D_full%rPE(1:com_elem2D_full%rPEnum)
  read(infile_cominfo,*) com_elem2D_full%rptr(1:com_elem2D_full%rPEnum+1)
  ALLOCATE(com_elem2D_full%rlist(eDim_elem2D+eXDim_elem2D))
  read(infile_cominfo,*) com_elem2D_full%rlist

  read(infile_cominfo,*) com_elem2D_full%sPEnum
  allocate(com_elem2D_full%sPE( com_elem2D_full%sPEnum))
  allocate(com_elem2D_full%sptr(com_elem2D_full%sPEnum+1))
  read(infile_cominfo,*) com_elem2D_full%sPE( 1:com_elem2D_full%sPEnum)
  read(infile_cominfo,*) com_elem2D_full%sptr(1:com_elem2D_full%sPEnum+1)
  n=com_elem2D_full%sptr(com_elem2D_full%sPEnum+1)-1
  ALLOCATE(com_elem2D_full%slist(n))
  read(infile_cominfo,*) com_elem2D_full%slist

  close(infile_cominfo)
  if (mype==0) print*,'communication arrays are read'
  
  call com_setup_haloexchange(com_nod2D)
  call com_setup_haloexchange(com_elem2D)
  call com_setup_haloexchange(com_elem2D_full)

  if(mype==0) print *,'Communication arrays are set' 
  
  ! copy the local dimension to the communicator struct
  com_nod2D%myDim       = myDim_nod2D
  com_nod2D%eDim        =  eDim_nod2D
  com_nod2D%exDim       =           0
  
  com_elem2D%myDim      = myDim_elem2D  
  com_elem2D%eDim       =  eDim_elem2D  
  com_elem2D%exDim      = exDim_elem2D

  com_elem2D_full%myDim = myDim_elem2D
  com_elem2D_full%eDim  =  eDim_elem2D
  com_elem2D_full%exDim = exDim_elem2D

  ! mype=0 gets knowledge about the domain decomposition on all other PEs.
  ! mype=0 will do the IO and needs the full global <-> local mapping
  call com_setup_scatterGatherLists

  
  
end subroutine com_read_subdomains
!========================================

subroutine com_printstruct(com, printarrays)

  type(com_struct), intent(in)  :: com
  logical, intent(in), optional :: printarrays
  logical                       :: allflag

! Debugging, debuggin...
  
  if (present(printarrays)) then
     allflag = printarrays
  else
     allflag = .false.
  end if

  print *, trim(com%name),' on', mype,'mpitype_maxnval', com%mpitype_maxnval
  print *, trim(com%name),' on', mype,'rPEnum',          com%rPEnum
  print *, trim(com%name),' on', mype,'rPE(:)',          com%rPE
  if (allflag) print *, trim(com%name),' on', mype,'rptr(:)',         com%rptr(:)      
  if (allflag) print *, trim(com%name),' on', mype,'rlist(:)',        com%rlist(:)     
  if (allflag) print *, trim(com%name),' on', mype,'rtype(:,:)',      com%rtype(:,:)   
  if (allflag) print *, trim(com%name),' on', mype,'rtype_i(:,:)',    com%rtype_i(:,:)
  print *, trim(com%name),' on', mype,'sPEnum',          com%sPEnum       
  print *, trim(com%name),' on', mype,'sPE(:)',          com%sPE(:)       
  if (allflag) print *, trim(com%name),' on', mype,'sptr(:)',         com%sptr(:)      
  if (allflag) print *, trim(com%name),' on', mype,'slist(:)',        com%slist(:)     
  if (allflag) print *, trim(com%name),' on', mype,'stype(:,:)',      com%stype(:,:)
  if (allflag) print *, trim(com%name),' on', mype,'stype_i(:,:)',    com%stype_i(:,:) 
  if (allflag) print *, trim(com%name),' on', mype,'req(:)',          com%req(:)       
  print *, trim(com%name),' on', mype,'nreq',            com%nreq         
  print *, trim(com%name),' on', mype,'myDim',           com%myDim        
  print *, trim(com%name),' on', mype,'globalDim',       com%globalDim    
  if (allflag) print *, trim(com%name),' on', mype,'mylist(:)',       com%mylist(:)    
  print *, trim(com%name),' on', mype,'gmax_len',        com%gmax_len     
  print *, trim(com%name),' on', mype,'gptr(:)',         com%gptr(:)      
  if (allflag) print *, trim(com%name),' on', mype,'glist(:)',        com%glist(:)     
  
end subroutine com_printstruct

#else
! Dummy subroutines for no MPI

  real(kind=dp) function walltime()
!$    use OMP_LIB
    integer(kind=int64) :: count, count_rate
!$    walltime = omp_get_wtime()

!$ if (.false.) then    
    call system_clock(count, count_rate)
    walltime = real(count,dp) / real(count_rate,dp)
!$ endif
  end function walltime

  subroutine com_parallel_init 
    ! nothing to be done    
  end subroutine com_parallel_init

  subroutine com_setup_haloexchange(com)
    integer, intent(in) :: com
    ! nothing to be done
  end subroutine com_setup_haloexchange

  subroutine com_setup_scatterGatherLists
  ! nothing to be done
  end subroutine com_setup_scatterGatherLists

  subroutine com_parallel_exit(ierr)

    integer, intent(in) :: ierr

    !NR: Problem: "stop ierr" with variable integer ierr is Fortran 2018,
    !NR: but not supported in ifort as of 19.1.2. Workaround:  
    if (ierr==0) then
       stop 0
    elseif (ierr /= 0) then
       print *,'Stopping with ierr=',ierr
       stop 1
    endif
  
  end subroutine com_parallel_exit
  
#endif

end module mpi_parallel

