!############################################################################
!#                                                                          #
!# Domain decomposition, stand alone code for TsunAWI meshes.               #
!# Basis taken from FESOM and FESOM-C domain decomposition.                 #
!#                                                                          #
!# Authors (with no claim to completeness): Sergey Danilov,                 #
!#   Stephan Frickenhaus, Vadym Aizinger, Sven Harig, Natalja Rakowsky,...  #
!############################################################################

MODULE  g_PARSUP

  integer, parameter  :: WP=4        ! Working precision
  CHARACTER(len=100)  :: MeshPath


! All variables used to keep the mesh structure +
! auxiliary variables involved in implementation 
! of open boundaries and advection schemes
! 

integer, parameter                         :: MAX_ADJACENT=32 ! Max allowed number of adjacent nodes
integer                                    ::   nod2D      ! the number of 2D nodes
real(kind=WP), allocatable, dimension(:,:) ::   coord_nod2D
integer                                    ::   edge2D     ! the number of 2D edges
integer                                    ::   elem2D     ! the number of 2D elements
integer, allocatable, dimension(:,:)       ::   elem2D_nodes
                                              ! elem2D_nodes(:,n) lists
				              ! 3-4 nodes of element n   
integer, allocatable, dimension(:,:)       ::   edges
                                              ! edge(:,n) lists 2 nodes
				              ! edge n
integer, allocatable, dimension(:,:)       ::   edge_tri
integer, allocatable, dimension(:,:)       ::   elem_edges
integer, allocatable, dimension(:,:)       ::   elem_neighbors
integer, allocatable, dimension(:,:)       ::   nod_in_elem2D
integer, allocatable, dimension(:)         ::   nod_in_elem2D_num
real(kind=WP), allocatable, dimension(:)   ::   depth


  type sparse_matrix 
     integer :: nza
     integer :: dim
     integer(KIND=4), allocatable,   dimension(:) :: colind
     integer(KIND=4), allocatable,   dimension(:) :: rowptr
  end type sparse_matrix
! Elevation stiffness matrix
type(sparse_matrix)                           :: ssh_stiff


 integer, parameter   :: MAX_LAENDERECK=8
 integer, parameter   :: MAX_NEIGHBOR_PARTITIONS=32
  type com_struct
     integer                                       :: rPEnum ! the number of PE I receive info from 
     integer, dimension(MAX_NEIGHBOR_PARTITIONS)   :: rPE    ! their list
     integer, dimension(MAX_NEIGHBOR_PARTITIONS+1) :: rptr   ! allocatables to the list of nodes
     integer, dimension(:), allocatable            :: rlist  ! the list of nodes
     integer                                       :: sPEnum ! send part 
     integer, dimension(MAX_NEIGHBOR_PARTITIONS)   :: sPE
     integer, dimension(MAX_NEIGHBOR_PARTITIONS)   :: sptr
     integer, dimension(:), allocatable            :: slist
  end type com_struct

  type(com_struct)   :: com_nod2D
  type(com_struct)   :: com_elem2D
  type(com_struct)   :: com_elem2D_full

  integer            :: npes
  integer            :: maxPEnum=100
  integer, allocatable, dimension(:)  :: part

  ! Mesh partition
  integer                             :: myDim_nod2D, eDim_nod2D
  integer, allocatable, dimension(:)  :: myList_nod2D
  integer                             :: myDim_elem2D, eDim_elem2D, eXDim_elem2D
  integer, allocatable, dimension(:)  :: myList_elem2D
  integer                             :: myDim_edge2D, eDim_edge2D
  integer, allocatable, dimension(:)  :: myList_edge2D


!$OMP threadprivate(com_nod2D,com_elem2D,com_elem2D_full)
!$OMP threadprivate(myDim_nod2D, eDim_nod2D, myList_nod2D)
!$OMP threadprivate(myDim_elem2D, eDim_elem2D, eXDim_elem2D, myList_elem2D)
!$OMP threadprivate(myDim_edge2D, eDim_edge2D, myList_edge2D)
  
end MODULE g_PARSUP



program MAIN

  use g_PARSUP

  implicit none
  character(len=1000)   :: nmlfile  !> name of configuration namelist file
  integer               :: istat, n
  character(len=10)     :: part_str
  

  integer              :: n_levels = 1    ! Number of levels for hierarchic partitioning
  integer, allocatable :: n_part(:)       ! Number of partitions on each hierarchy level

    CHARACTER(len=100)  :: IniCondPath
    CHARACTER(len=100)  :: OutputPath
    CHARACTER(len=100)  :: TopoFile
    CHARACTER(len=100)  :: TideGaugeFile
    CHARACTER(len=100)  :: output_prefix
    INTEGER             :: coordinate_type 
    INTEGER             :: rotation_type
    INTEGER             :: momadv_type
    INTEGER             :: wetting_drying_type
    LOGICAL             :: use_bv_roughness
    INTEGER             :: nmb_iter_smooth_inicond
    INTEGER             :: nmb_iter_smooth_topo
    INTEGER             :: verbosity



    NAMELIST /general/ MeshPath, IniCondPath, OutputPath, TopoFile, TideGaugeFile, output_prefix,  &
        coordinate_type, rotation_type, momadv_type, wetting_drying_type, &
        use_bv_roughness,  &
        nmb_iter_smooth_inicond, nmb_iter_smooth_topo, verbosity
    
  nmlfile ='./namelist.tsunami'
  open (20,file=nmlfile)
  read (20,NML=general)         ! We need MeshPath
  close (20)

  n_levels = COMMAND_ARGUMENT_COUNT()

  if (n_levels==0) then
     print *,'please give number of subdomains as input parameter(s)'
     stop 1
  end if
  allocate(n_part(n_levels))
  do n=1,n_levels
     call GET_COMMAND_ARGUMENT(n, part_str, status=istat)
     if (istat==0) read(part_str,*,iostat=istat) n_part(n) 
     if (istat /= 0) then
        print *,'error',istat,'reading number of subdomains on level', n
        stop 1
     elseif (n_part(n) < 2) then
        print *,'partitions on each level must be >= 2'
        stop 1
     end if
  enddo
      
  call read_mesh_ini
  call find_edges_ini
  call find_elem_neighbors_ini
! NR Some arrays are not needed for partitioning, after setting up the grid
  deallocate(coord_nod2D)
  call stiff_mat_ini
  call set_par_support_ini(n_levels,n_part)
  call communication_ini

end program MAIN
!=============================================================================
!> @brief
!> Reads mesh files 
subroutine read_mesh_ini

USE g_PARSUP

!
IMPLICIT NONE
!
INTEGER               :: nq
INTEGER               :: n1,n2,n3, n4
INTEGER               :: n, nz, exit_flag, el
REAL(kind=WP)         :: x1, x2
INTEGER	              :: tag
INTEGER               :: i_error, i_err
  character(len=200) :: a_dummy

! ===================
! Surface mesh
! ===================
  open (20,file=trim(meshpath)//'nod2d.out', status='old')
  READ(20,*) nod2D
  ALLOCATE(coord_nod2D(2,nod2D))
    
  ! Test, if nod2d.out contains 3 or 4 columns.
  ! The first column, the number of the node, may be skipped.
  READ(20,'(a)') a_dummy
  READ(a_dummy,*,IOSTAT=i_err) nq, x1, x2, tag

  if (i_err==0) then
     coord_nod2D(1,nq) = x1
     coord_nod2D(2,nq) = x2
     DO n=2, nod2D
        READ(20, *) nq, x1, x2, tag
        coord_nod2D(1,nq) = x1
        coord_nod2D(2,nq) = x2
     END DO
  else
     READ(a_dummy,*) coord_nod2D(1:2,1)
     DO n=2, nod2D
        READ(20, *) coord_nod2D(1:2,n)
     end DO
  end if
  CLOSE(20)

  
  open (21,file=trim(meshpath)//'elem2d.out', status='old')
  READ(21,*)  elem2D
    
  ! Test, if elem2d.out contains 3 or 4  columns.
  
  ALLOCATE(elem2D_nodes(4,elem2D))
  READ(21,'(a)') a_dummy
  READ(a_dummy,*,IOSTAT=i_err) n1, n2, n3, n4
  
  if (i_err==0) then
     ! 4 columns, thus a quad or mixed mesh
     elem2D_nodes(1:4,1) = (/ n1, n2, n3, n4 /)
     DO el = 2, elem2D
        read(21,*) elem2D_nodes(1:4,el)
     END DO
  else
     ! 3 columns, triangular mesh
     read(a_dummy, *) elem2D_nodes(1:3,1)
     elem2D_nodes(4,1) = elem2D_nodes(1,1)
     DO el = 2, elem2D
        read(21,*) elem2D_nodes(1:3,el)
        elem2D_nodes(4,el) = elem2D_nodes(1,el)
     END DO
  end if
  !
  CLOSE(21)

   write(*,*) '========================='
   write(*,*) 'Mesh is read : ', 'nod2D=', nod2D,' elem2D=', elem2D
   write(*,*) '========================='

END SUBROUTINE read_mesh_ini

!===================================================================
! Stiffness matrix for the elevation
subroutine stiff_mat_ini
  use g_PARSUP
  
  !
  implicit none
  integer                :: i, j, n, q, el, elem_nodes_max, nod(4)
  integer, allocatable   :: num_ne(:), ne(:,:)
  !

  ssh_stiff%dim = nod2D   
  allocate(ssh_stiff%rowptr(nod2D+1))

  allocate(num_ne(nod2D), ne(MAX_ADJACENT,nod2D))
  num_ne(:)           = 0
  ne(:,:)             = -1
 
  ! Check the maximum number of nodes in an element (FESOM triangular meshes = 3, Hybrid meshes = 4)
  elem_nodes_max = size(elem2D_nodes, 1)

  ! Fill node adjacency info
  ! all nodes in an element are adjacent in the sense of being halo nodes
  ! (also the opposite nodes of a quad: no edge, but the indirect connection
  ! should be taken into account by metis domain decomposition)
  do el=1,elem2D
     nod(1:elem_nodes_max) = elem2D_nodes(1:elem_nodes_max,el)  ! Fortran-numbering
     q = elem_nodes_max
     if (nod(1) == nod(elem_nodes_max)) q = q-1  ! triangle
    
     do i=2,q
        do j=1,i-1
           if (all(ne(:,nod(i)) /= nod(j))) then
              num_ne(nod(i)) = num_ne(nod(i)) + 1
              num_ne(nod(j)) = num_ne(nod(j)) + 1

              if (max(num_ne(nod(i)), num_ne(nod(j))) > MAX_ADJACENT ) then
                 print *,'Recompile with larger value for parameter MAX_ADJACENT.'
                 stop
              else
                 ne(num_ne(nod(i)), nod(i)) = nod(j)
                 ne(num_ne(nod(j)), nod(j)) = nod(i)
              endif
           endif
        end do
     end do
  end do

! copy adjacency matrix to CSR-format
  ssh_stiff%rowptr(1) = 1
  do n=1,nod2D
     ssh_stiff%rowptr(n+1) = ssh_stiff%rowptr(n)+num_ne(n)
  end do
  
  allocate(ssh_stiff%colind(ssh_stiff%rowptr(nod2D+1)-1))  
  do n=1,nod2D
     ssh_stiff%colind(ssh_stiff%rowptr(n):ssh_stiff%rowptr(n+1)-1) = ne(1:num_ne(n),n)
  end do

  deallocate(num_ne, ne)

end subroutine stiff_mat_ini

!===================================================================
! Setup of communication arrays
subroutine communication_ini

  USE g_PARSUP
!$  use omp_lib
  implicit none

  integer        :: n, pe
  character*10   :: npes_string
  character*200  :: dist_mesh_dir
  LOGICAL        :: L_EXISTS

  ! Create the distributed mesh subdirectory
  write(npes_string,"(I10)") npes
  dist_mesh_dir=trim(meshpath)//'dist_'//trim(ADJUSTL(npes_string))//'/'
  INQUIRE(file=trim(dist_mesh_dir), EXIST=L_EXISTS)
  if (.not. L_EXISTS) call system('mkdir '//trim(dist_mesh_dir))

#ifdef OMP_MAX_THREADS
!$OMP PARALLEL NUM_THREADS(OMP_MAX_THREADS)
  if (OMP_GET_THREAD_NUM() == 0) then
     write(*,*) 'Setting up communication arrays using ', OMP_GET_NUM_THREADS(), ' threads'
  endif
#else
!$OMP PARALLEL NUM_THREADS(1)
  write(*,*) 'Setting up communication arrays using 1 thread (serially)'
#endif
  
!$OMP DO
  do pe = 0, npes-1
     call communication_nodn(pe)
     call communication_elemn(pe)

     call save_dist_mesh(pe)         ! Write out communication file com_infoxxxxx.out
  end do
!$OMP END DO
!$OMP END PARALLEL

  deallocate(elem_neighbors)
  deallocate(part)
  write(*,*) 'Communication arrays have been set up'   
end subroutine communication_ini


!=======================================================================
SUBROUTINE find_elem_neighbors_ini
! For each element three its element neighbors are found

USE g_PARSUP
implicit none
integer    :: elem, j, n, elnodes(4), el, edg, q

allocate(elem_neighbors(4,elem2D))

DO el=1,elem2D
   
   DO j=1,4
      edg = elem_edges(j,el)
      if (edg == 0) then
         ! for j=4, triangle
         elem_neighbors(j,el) = 0
      elseif(edge_tri(1,edg) /= el) then
         elem_neighbors(j,el) = edge_tri(1,edg)
      else
         elem_neighbors(j,el) = edge_tri(2,edg)
      end if
   END DO
   
END DO

 deallocate(edge_tri)
 

 ! Among elem_neighbors there can be negative numbers. These correspond to 
 ! boundary elements for which neighbours are absent.


 ! =============
 ! To facilitate computations the neibourhood
 ! information is assembled
 ! =============	 
 allocate(nod_in_elem2D_num(nod2D))
 nod_in_elem2D_num=0
 do n=1,elem2D
    elnodes=elem2D_nodes(:,n)
    if (elnodes(4)==elnodes(1)) then !triangle
       nod_in_elem2D_num(elnodes(1:3))=nod_in_elem2D_num(elnodes(1:3))+1
    else ! quad
       nod_in_elem2D_num(elnodes)=nod_in_elem2D_num(elnodes)+1
    end if
 end do
 allocate(nod_in_elem2D(maxval(nod_in_elem2D_num),nod2D))
 nod_in_elem2D=0
 
 nod_in_elem2D_num=0
 do n=1,elem2D   
    elnodes=elem2D_nodes(:,n)
    if (elnodes(4)==elnodes(1)) then !triangle
       nod_in_elem2D_num(elnodes(1:3))=nod_in_elem2D_num(elnodes(1:3))+1
    else !quad
       nod_in_elem2D_num(elnodes)=nod_in_elem2D_num(elnodes)+1
    end if
    do j=1,3
       nod_in_elem2D(nod_in_elem2D_num(elnodes(j)),elnodes(j))=n
    end do
    if (elnodes(4)/=elnodes(1)) nod_in_elem2D(nod_in_elem2D_num(elnodes(4)),elnodes(4))=n
 end do  

END SUBROUTINE find_elem_neighbors_ini
!=========================================================================
!> @brief
!> Finds edges. Creates 3 files: edgenum.out, edges.out, edge_tri.out
SUBROUTINE find_edges_ini

USE g_PARSUP

IMPLICIT NONE
integer, allocatable                  :: aux1(:), ne_num(:), ne_pos(:,:)
integer, allocatable                  ::          nn_num(:), nn_pos(:,:)
integer                               :: counter, n, k, q, edg, el
integer                               :: elem, elem1, elems(2), q1, q2
integer                               :: elnodes(4), ed(2), flag, eledges(4)
integer                               :: temp(100), node 
real(kind=WP)                         :: xc(2), xe(2), ax(3), amin

! ====================
! (a) find edges. To make the procedure fast 
! one needs neighbourhood arrays
! ====================

allocate(ne_num(nod2d), ne_pos(MAX_ADJACENT, nod2D), nn_num(nod2D))
ne_num=0
DO el=1,elem2D
    elnodes = elem2D_nodes(:,el)
    q1=4
    if(elnodes(1)==elnodes(4)) q1=3
    DO q = 1,q1
       ne_num(elnodes(q)) = ne_num(elnodes(q))+1
       if (ne_num(elnodes(q)) > MAX_ADJACENT ) then
          print *,'Recompile with larger value for parameter MAX_ADJACENT.'
          stop
       else
          ne_pos(ne_num(elnodes(q)),elnodes(q)) = el
       endif
    END Do
END DO                 ! neighbor elements are found 

! count neighbour nodes				     
! In quads we should count the nodes that are 
! connected by edges!
allocate(aux1(nod2D))                   
aux1=0
 
DO n=1, nod2D
    counter=0
    DO k=1, ne_num(n)
        elem=ne_pos(k,n)
        elnodes=elem2D_nodes(:,elem)        
        if(elnodes(1)==elnodes(4)) then
            DO q=1,3
                if(elnodes(q)==n) CYCLE 
                if(aux1(elnodes(q)) /= 1) then
                    counter=counter+1
                    aux1(elnodes(q))=1
                    temp(counter)=elnodes(q)
                end if
            END DO
        else
        ! Find the position of n in elnodes:
            if(elnodes(1)==n .or. elnodes(3)==n) then
                ed(1)=elnodes(2)
                ed(2)=elnodes(4)
            else
                ed(1)=elnodes(1)
                ed(2)=elnodes(3)
            end if
            DO q=1,2
                if(aux1(ed(q)) /= 1) then
                    counter=counter+1
                    aux1(ed(q))=1
                    temp(counter)=ed(q)
                end if
            END DO
        end if
    END DO
    nn_num(n)=counter         
    aux1(temp(1:counter))=0
END DO              
 
allocate(nn_pos(maxval(nn_num)+1,nod2D))
nn_pos = -1
aux1=0
 
DO n=1, nod2D
    counter=0
    DO k=1, ne_num(n)
        elem=ne_pos(k,n)
        elnodes=elem2D_nodes(:,elem)        
        if(elnodes(1)==elnodes(4)) then 
            DO q=1,3
                if(elnodes(q)==n) CYCLE 
                if(aux1(elnodes(q)) /= 1) then
                    counter=counter+1
                    aux1(elnodes(q))=1
                    temp(counter)=elnodes(q)
                end if
            END DO
        else
            ! Find the position of n in elnodes:
            if(elnodes(1)==n .or. elnodes(3)==n) then
                ed(1)=elnodes(2)
                ed(2)=elnodes(4)
            else
                ed(1)=elnodes(1)
                ed(2)=elnodes(3)
            end if
            DO q=1,2
                if(aux1(ed(q)) /= 1) then
                    counter=counter+1
                    aux1(ed(q))=1
                    temp(counter)=ed(q)
                end if
            END DO
        end if
    END DO
    nn_num(n)             = counter+1
    aux1(temp(1:counter)) = 0
    nn_pos(1,n)           = n 
    nn_pos(2:counter+1,n) = temp(1:counter)
END DO              
deallocate(aux1)
! neighboring nodes are found. First in the list is the node itself 

! ====================
! (b) Find edges and elements containing them.
! ====================

 ! Count edges: 
 ! ==================== 
 ! form edges with n by cycling over neighbor
 ! nodes (if edges are not accounted yet). 
 ! New edges are added only if neighbor>n  
 ! ====================

 edge2D = (sum(nn_num(1:nod2D))-nod2D)/2
 
 counter = 0
 allocate(edges(2,edge2D), edge_tri(2, edge2D))
 DO n = 1,nod2D
    DO q = 2,nn_num(n)
       node = nn_pos(q,n)
       if(node < n) CYCLE
       counter = counter+1   ! new edge (n,node)
       edges(1,counter)    = n
       edges(2,counter)    = node
       
       ! find elements containing n and node
       flag=0
       DO k=1, ne_num(n)
          el = ne_pos(k,n)
          if (any(elem2D_nodes(1:4,el) == node)) then
             flag = flag + 1
             edge_tri(flag,counter) = el
             if (flag == 2) exit
          endif
       END DO
       
    END DO
 END DO


 ! ====================
 ! (e) We need an array inverse to edge_tri listing edges
 ! of a given triangle 
 ! ====================
 allocate(elem_edges(4,elem2D))
 elem_edges=0

 DO edg=1, edge2D
    DO k=1,2
       el=edge_tri(k,edg)   ! triangle number
       if (el>0) then
	  if (elem_edges(1,el) == 0) then
             elem_edges(1,el) = edg
	  elseif (elem_edges(2,el) == 0) then
             elem_edges(2,el) = edg
	  elseif (elem_edges(3,el) == 0) then
             elem_edges(3,el) = edg
	  elseif (elem_edges(4,el) == 0) then
             elem_edges(4,el) = edg
          else
             print *,'problem with quad', el
             stop
          end if
          end if
    END DO
 END DO

 deallocate(ne_num, ne_pos)
END SUBROUTINE find_edges_ini

!=================================================================
subroutine set_par_support_ini(n_levels,n_part)

  use iso_c_binding, only: idx_t=>C_INT32_T
  use g_PARSUP
  
  implicit none
  
  
    integer, intent(in) :: n_levels, n_part(n_levels)
    integer        :: n, j, k, nini, nend, ierr
    integer(idx_t) :: np(n_levels+1)
    integer(idx_t) :: weights(nod2D)
  
  
  interface 
     subroutine partit(n,ptr,adj,wgt,np,part) bind(C)
       use iso_c_binding, only: idx_t=>C_INT32_T
       integer(idx_t), intent(in)  :: n, ptr(*), adj(*), wgt(*), np(*)
       integer(idx_t), intent(out) :: part(*)
     end subroutine partit
  end interface

  np(1:n_levels) = n_part(:)               ! Number of partitions on each hierarchy level
  np(n_levels+1) = 0 
  

  allocate(part(nod2D))
  part=0

  npes = PRODUCT(np(1:n_levels))
  if(npes<2) then
     print *,'Total number of parallel partitions is less than one! Aborting...'
     stop
  end if


  weights(1:nod2D) = 0  ! No additional weight so far
  
  write(*,*) 'Calling partit'
  call partit(ssh_stiff%dim, ssh_stiff%rowptr, ssh_stiff%colind, &
       weights, np, part)

  call check_partitioning

  write(*,*) 'Partitioning is done.'

! The stiffness matrix is no longer needed. 
  deallocate(ssh_stiff%rowptr)
  deallocate(ssh_stiff%colind)

end subroutine set_par_support_ini

!=======================================================================

subroutine check_partitioning

  ! In general, METIS 5 has several advantages compared to METIS 4, e.g.,
  !   * neighbouring tasks get neighbouring partitions (important for multicore computers!)
  !   * lower maximum of weights per partition (better load balancing)
  !   * lower memory demand
  !
  ! BUT: there might be outliers, single nodes connected to their partition by
  !      only one edge or even completely isolated. This spoils everything :-(
  !
  ! This routine checks for isolated nodes and moves them to an adjacent partition,
  ! trying not to spoil the load balance.

  use g_PARSUP
  
  integer :: i, j, k, n, n_iso, n_iter, is, ie, kmax, np, cnt
  integer :: nod_per_partition(0:npes-1)
  integer :: max_nod_per_part, min_nod_per_part
  integer :: average_nod_per_part, node_neighb_part(100)
  logical :: already_counted, found_part

  integer :: max_adjacent_nodes
  integer, allocatable :: ne_part(:), ne_part_num(:), ne_part_load(:)

  ! Check load balancing
  do i=0,npes-1
     nod_per_partition(i) = count(part(:) == i)
  enddo

  min_nod_per_part = minval( nod_per_partition(:))

  max_nod_per_part = maxval( nod_per_partition(:))

  average_nod_per_part = nod2D / npes

  ! Now check for isolated nodes (connect by one or even no edge to other
  ! nodes of its partition) and repair, if possible

  max_adjacent_nodes = maxval(ssh_stiff%rowptr(2:nod2D+1) - ssh_stiff%rowptr(1:nod2D))
  allocate(ne_part(max_adjacent_nodes), ne_part_num(max_adjacent_nodes), &
           ne_part_load(max_adjacent_nodes))

  isolated_nodes_check: do n_iter = 1, 10
     print *,' '
     print *,'Check for isolated nodes, new iteration ========'
     n_iso = 0
     do n=1,nod2D
        is = ssh_stiff%rowptr(n)
        ie = ssh_stiff%rowptr(n+1) -1
        cnt = ie-is+1 
        
        node_neighb_part(1:cnt) = part(ssh_stiff%colind(is:ie))
        if (count(node_neighb_part(1:cnt) == part(n)) <= 1) then

           n_iso = n_iso+1
           print *,'Isolated node',n, 'in partition', part(n)
           print *,'Neighbouring nodes are in partitions',  node_neighb_part(1:cnt)
           print *,'Nodes:',  ssh_stiff%colind(is:ie)

           ! count the adjacent nodes of the other PEs

           np=1
           ne_part = node_neighb_part
           ne_part_num = 1
           ne_part_load(1) = nod_per_partition(ne_part(1)) + 1

           do i=1,cnt
              if (node_neighb_part(i)==part(n)) cycle
              already_counted = .false.
              do k=1,np
                 if (node_neighb_part(i) == ne_part(k)) then
                    ne_part_num(k) = ne_part_num(k) + 1
                    already_counted = .true.
                    exit
                 endif
              enddo
              if (.not. already_counted) then
                 np = np+1
                 ne_part(np) = node_neighb_part(i)
                 ne_part_num(np) = 1
                 ne_part_load(np) = nod_per_partition(ne_part(np)) + 1
              endif
           enddo

           ! Now, check for two things: The load balance, and if 
           ! there is more than one node of that partition.
           ! Otherwise, it would become isolated again.

           ! Best choice would be the partition with most adjacent nodes (edgecut!)
           ! Choose, if it does not decrease the load balance. 
           !        (There might be two partitions with the same number of adjacent
           !         nodes. Don't care about this here)

           kmax = maxloc(ne_part_num(1:np),1)

           if (ne_part_num(kmax) <= 1) then
              print *,'Sorry, no chance to solve an isolated node problem'
              exit isolated_nodes_check
           endif

           if  (ne_part_load(kmax) <= max_nod_per_part ) then
              k = kmax
           else
              ! Don't make it too compicated. Reject partitions that have only one
              ! adjacent node. Take the next not violating the load balance.
              found_part = .false.
              do k=1,np
                 if (ne_part_num(k)==1 .or. k==kmax) cycle

                 if  (ne_part_load(k) <= max_nod_per_part ) then

                    found_part = .true.
                    exit
                 endif
              enddo

              if (.not. found_part) then
                 ! Ok, don't think to much. Simply go for minimized edge cut.
                 k = kmax
              endif
           endif

           ! Adjust the load balancing

           nod_per_partition(ne_part(k)) = nod_per_partition(ne_part(k)) + 1 
           nod_per_partition(part(n))    = nod_per_partition(part(n)) - 1

           ! And, finally, move nod n to other partition        
           part(n) = ne_part(k)
           print *,'Node',n,'is moved to part',part(n)
        endif
     enddo

     if (n_iso==0) then
        print *,'No isolated nodes found'
        exit isolated_nodes_check 
     endif
     ! Check for isolated nodes again
  end do isolated_nodes_check

  deallocate(ne_part, ne_part_num, ne_part_load)

  if (n_iso > 0) then
     print *,'++++++++++++++++++++++++++++++++++++++++++++'
     print *,'+  WARNING: PARTITIONING LOOKS REALLY BAD  +'
     print *,'+  It was not possible to remove all       +'
     print *,'+  isolated nodes. Consider to rerun with  +'
     print *,'+  new metis random seed                   +'
     print *,'++++++++++++++++++++++++++++++++++++++++++++'
  endif

  print *,'=== LOAD BALANCING ==='
  print *,'2D nodes: min, aver, max per part',min_nod_per_part, &
       average_nod_per_part,max_nod_per_part

  write(*,"('2D nodes: percent min, aver, max ',f8.3,'%, 100%, ',f8.3,'%')") &
       100.*real(min_nod_per_part) / real(average_nod_per_part), &
       100.*real(max_nod_per_part) / real(average_nod_per_part)


end subroutine check_partitioning

!=======================================================================
subroutine communication_nodn(mype)

  use g_PARSUP
  implicit none

  integer, intent(in)    :: mype
  integer                :: n, np, prank, el, r_count, s_count, q, i, j, nod, k, l
  integer                :: num_send(0:npes-1), num_recv(0:npes-1), nd_count
  integer, allocatable   :: recv_from_pe(:), send_to_pes(:,:)
  logical                :: max_laendereck_too_small=.false.
  integer                :: IERR

  ! Assume we have 2D partitioning vector in part. Find communication rules
  ! Reduce allocation: find all neighboring PE

  nd_count = count(part(1:nod2d) == mype)
  allocate(recv_from_pe(nod2d), send_to_pes(MAX_LAENDERECK,nd_count), &
       myList_nod2D(nd_count), STAT=IERR)
  if (IERR /= 0) then
     write (*,*) 'Could not allocate arrays in communication_nodn'
     stop
  endif

  nd_count = 0
  do n=1,nod2D
     ! Checks if element el has nodes that belong to different partitions
     if (part(n) == mype) then
        nd_count = nd_count+1
        myList_nod2D(nd_count)=n
     endif
  end do
  myDim_nod2D=nd_count

  num_send(0:npes-1) = 0
  num_recv(0:npes-1) = 0
  recv_from_pe(1:nod2d) = -1
  send_to_pes(1:MAX_LAENDERECK,1:nd_count) = -1

  ! For the local nodes, run through the patch and collect all nodes in the patch
  ! (It would be simpler to re-use the adjacency matrix, but it is not a global
  !  variable... and I am lazy and want to keep the data structure simple)

  do l=1,nd_count
     n = myList_nod2D(l)
     do i = 1, nod_in_elem2D_num(n)
        ! Over all elements el that the node n is part of
        el = nod_in_elem2D(i,n)
if (el==0) then
print *,'HEYA:',l,nd_count,n,i,nod_in_elem2D_num(n)
end if

        ! Checks, if elements are quads or triangles
        q = 4    ! quads as default
        if (elem2D_nodes(1,el) == elem2D_nodes(4,el)) q=3  ! triangle

        do j = 1, q
           ! Over all nodes in every element el
           nod = elem2D_nodes(j,el)
           
           ! Checks, if node j is not in another partitionen
           if (part(nod) /= mype) then
              ! Checks, if not already considered to be received from this
              ! node from partition part(nod)
              if (recv_from_pe(nod) == -1) then  ! nod already collected to be received?
                 ! We have to receive this node from part(nod)
                 ! Add plus one node to the total number of
                 num_recv(part(nod)) = num_recv(part(nod)) + 1
                 ! ???
                 recv_from_pe(nod) = part(nod)   ! recv_from_pe(recv_count) = nod  ! no new information, just handy
              endif
              ! Checks, if all possible connected partition
              ! And we have to send n to part(nod). Do we know this already?
              do k=1,MAX_LAENDERECK    !???
                 if (send_to_pes(k,l) == part(nod)) then
                    exit  ! already collected
                 elseif (send_to_pes(k,l) == -1) then
                    send_to_pes(k,l) = part(nod)
                    num_send(part(nod)) = num_send(part(nod)) + 1
                    exit
                 elseif (k== MAX_LAENDERECK) then
                    max_laendereck_too_small = .true.  ! Problem
                 endif
              enddo
           endif
        enddo
     enddo
  enddo

  if (max_laendereck_too_small) then
     print *,'Increase MAX_LAENDERECK in gen_modules_partitioning.F90 and recompile'
     stop
  endif

  ! Now, build the send and recv communication data structure
  ! To how many PE needs to be send and which PEs
!!$  com_nod2D%rPEnum = count(num_recv(0:npes-1) > 0)
!!$  com_nod2D%sPEnum = count(num_send(0:npes-1) > 0)

!!$  if (com_nod2D%rPEnum > MAX_NEIGHBOR_PARTITIONS .or.  &
!!$       com_nod2D%sPEnum > MAX_NEIGHBOR_PARTITIONS) then
!!$     print *,'Increase MAX_NEIGHBOR_PARTITIONS in gen_modules_partitioning.F90 and recompile'
!!$     stop
!!$  endif
!!$  allocate(com_nod2D%rPE(com_nod2D%rPEnum))
!!$  allocate(com_nod2D%sPE(com_nod2D%sPEnum))

  r_count = 0
  s_count = 0
  com_nod2D%rptr(1) = 1
  com_nod2D%sptr(1) = 1

  do np = 0, npes-1
     if(num_recv(np) /= 0) then
        r_count = r_count+1
        com_nod2D%rPE(r_count) = np
        com_nod2D%rptr(r_count+1) =  com_nod2D%rptr(r_count)+ num_recv(np)
     end if
     if(num_send(np) /= 0) then
        s_count = s_count+1
        com_nod2D%sPE(s_count) = np
        com_nod2D%sptr(s_count+1) =  com_nod2D%sptr(s_count)+ num_send(np)
     end if
  enddo
  com_nod2D%rPEnum = r_count
  com_nod2D%sPEnum = s_count
  if (com_nod2D%rPEnum > MAX_NEIGHBOR_PARTITIONS .or.  &
       com_nod2D%sPEnum > MAX_NEIGHBOR_PARTITIONS) then
     print *,'Increase MAX_NEIGHBOR_PARTITIONS in gen_modules_partitioning.F90 and recompile'
     stop
  endif

  ! Counts the number of node for each partition PE mype has to send/receive
  ! In ascending order of PE number
!!$  r_count = 0
!!$  s_count = 0
!!$  allocate(com_nod2D%rptr(com_nod2D%rPEnum+1)) 
!!$  allocate(com_nod2D%sptr(com_nod2D%sPEnum+1))

!!$  com_nod2D%rptr(1) = 1
!!$  com_nod2D%sptr(1) = 1
!!$
!!$  do r_count = 1, com_nod2D%rPEnum
!!$     np = com_nod2D%rPE(r_count)
!!$     com_nod2D%rptr(r_count+1) =  com_nod2D%rptr(r_count)+ num_recv(np)
!!$  enddo
!!$  do s_count = 1, com_nod2D%sPEnum
!!$     np = com_nod2D%sPE(s_count)
!!$     com_nod2D%sptr(s_count+1) =  com_nod2D%sptr(s_count)+ num_send(np)
!!$  enddo

  ! Lists themselves

  r_count = 0
  eDim_nod2D=com_nod2D%rptr(com_nod2D%rPEnum+1)-1   
  allocate(com_nod2D%rlist(eDim_nod2D), &
       com_nod2D%slist(com_nod2D%sptr(com_nod2D%sPEnum+1)-1), STAT=IERR) 
  if (IERR /= 0) then
     write (*,*) 'Could not allocate arrays in communication_nodn'
     stop
  endif

  do np = 1,com_nod2D%rPEnum
     prank = com_nod2D%rPE(np)
     do n = 1, nod2D
        if (recv_from_pe(n) == prank) then
           r_count = r_count+1
           com_nod2D%rlist(r_count) = n
        end if
     end do
  end do

  s_count = 0
!!$  allocate(com_nod2D%slist(com_nod2D%sptr(com_nod2D%sPEnum+1)-1)) 
  do np = 1,com_nod2D%sPEnum
     prank = com_nod2D%sPE(np)
     do l = 1, nd_count
        n = myList_nod2D(l)
        if(any(send_to_pes(:,l) == prank)) then 
           s_count = s_count+1
           com_nod2D%slist(s_count) = n
        end if
     end do
  end do

  ! Summary of this piece: mype receives
  ! information on external 2D nodes from
  ! comm_nod2D%rPEnum external PEs
  ! Their ranks (numbers) are in array
  ! comm_nod2D%rPE(:)
  ! Pointers to external node numbers are in
  ! comm_nod2D%rptr(:)
  ! The node numbers are in 
  ! comm_nod2D%list(:)
  ! Putting everything into structure takes many operations, but
  ! without the structure we will need to many names and arrays
  ! Do not forget that we need also send part.

  ! mype sends its data to
  ! comm_nod2D%sPEnum external PEs
  ! Their ranks (numbers) are in array
  ! comm_nod2D%sPE(:)
  ! Pointers to external node numbers are in
  ! comm_nod2D%sptr(:)
  ! The node numbers are in 
  ! comm_nod2D%list(:)

  deallocate(recv_from_pe, send_to_pes)

   end subroutine communication_nodn

!==========================================================================
subroutine communication_elemn(mype)
 
  use g_PARSUP
  implicit none

  integer, intent(in)   :: mype
  integer, allocatable  :: recv_from_pe(:), send_to_pes(:,:)
  logical               :: max_laendereck_too_small=.false.
  integer               :: n, k, ep, np, prank, el, nod
  integer               :: p, q, j, elem, i, l, r_count, s_count, el_count
  integer               :: num_send(0:npes-1), num_recv(0:npes-1)
  integer               :: IERR

  
  ! Assume we have 2D partitioning vector in part. Find communication
  ! rules. An elem is external to element n if neither of its nodes 
  ! belongs to PE, but it is among the neighbors. Element n belongs to PE if 
  ! any of its nodes does. 
  
  ! This routine takes into account 
  ! com_elem2D_full: all  neighbors  
  ! com_elem2D:      only those sharing an edge 


  !===========================================
  !  com_elem2D
  !===========================================
  
  allocate(recv_from_pe(elem2D), STAT=IERR)
  if (IERR /= 0) then
     write (*,*) 'Could not allocate arrays in communication_elemn'
     stop
  endif

  el_count = 0
  do el=1,elem2D
     ! Checks if element el has nodes that belong to different partitions
     if (any(part(elem2D_nodes(1:4,el)) == mype)) then
        el_count = el_count+1
        recv_from_pe(el_count) = el
     endif
  end do
  myDim_elem2D=el_count

  allocate(myList_elem2D(el_count), send_to_pes(MAX_LAENDERECK,el_count), STAT=IERR)
  if (IERR /= 0) then
     write (*,*) 'Could not allocate arrays in communication_elemn'
     stop
  endif

  myList_elem2D(1:el_count) = recv_from_pe(1:el_count)
  num_send(0:npes-1) = 0
  num_recv(0:npes-1) = 0
  recv_from_pe(1:elem2D) = -1
  send_to_pes(1:MAX_LAENDERECK,1:el_count) = -1
  
  ! For the local elements, collect all adjacent (sharing an edge) elements
  ! that belong to other PEs
  
  do l=1,el_count
     el = myList_elem2D(l)

     q=4
     if (elem2D_nodes(1,el) == elem2D_nodes(4,el)) q=3 ! Triangle

     do n = 1, q                       ! cycle through all nodes

        ! Neighboring element elem that shares an edge with element el
        elem = elem_neighbors(n,el)

        if (elem < 1) cycle  ! boundary, "ghost element"

        ! Check for elements to be received
        if (all(part(elem2D_nodes(1:4,elem)) /= mype) .and. recv_from_pe(elem)==-1) then
           ! elem to be received already collected?
           ! We have to receive elem from PE ep:
           ep = part(elem2D_nodes(1,elem))  ! PE of first node is "main" owner
           num_recv(ep) = num_recv(ep) + 1
           recv_from_pe(elem) = ep  
        endif

        ! Check for elements to be sent to
        ! And maybe, we have to send el to the owners of elem
        ! 1. Is partition mype the main owner of element el?
        if (part(elem2D_nodes(1,el)) == mype .and. any(part(elem2D_nodes(1:4,elem)) /= mype)) then

           ! 2. Who owns element elem and needs to get element el? We must check all nodes!
           p=merge(3, 4, elem2D_nodes(1,elem) == elem2D_nodes(4,elem))

           do i=1,p
              ep = part(elem2D_nodes(i,elem))

              ! 3. Is ep also an owner of el and no send is needed? This excludes also mype==ep.
              if (any(part(elem2D_nodes(1:q,el)) == ep)) cycle

              ! 4. Ok, for the owner ep, check if sending el is already collected
              do k=1,MAX_LAENDERECK
                 if (send_to_pes(k,l) == ep) then
                    exit  ! already collected
                 elseif (send_to_pes(k,l) == -1) then
                    send_to_pes(k,l) = ep
                    num_send(ep) = num_send(ep) + 1
                    exit
                 elseif (k== MAX_LAENDERECK) then
                    max_laendereck_too_small = .true.  ! Problem
                 endif
              enddo
           enddo
        end if
     end do
  enddo

  if (max_laendereck_too_small) then
     print *,'Increase MAX_LAENDERECK in gen_modules_partitioning.F90 and recompile'
     stop
  endif
    
! Now, build the send and recv communication data structure
!!$  com_elem2D%rPEnum = count(num_recv(0:npes-1) > 0)
!!$  com_elem2D%sPEnum = count(num_send(0:npes-1) > 0)

!!$  if (com_elem2D%rPEnum > MAX_NEIGHBOR_PARTITIONS .or.  &
!!$       com_elem2D%sPEnum > MAX_NEIGHBOR_PARTITIONS) then
!!$     print *,'Increase MAX_NEIGHBOR_PARTITIONS in gen_modules_partitioning.F90 and recompile'
!!$     stop
!!$  endif
!!$  allocate(com_elem2D%rPE(com_elem2D%rPEnum))
!!$  allocate(com_elem2D%sPE(com_elem2D%sPEnum))

  r_count = 0
  s_count = 0
  com_elem2D%rptr(1) = 1
  com_elem2D%sptr(1) = 1

  do np = 0, npes-1
     if(num_recv(np) /= 0) then
        r_count = r_count+1
        com_elem2D%rPE(r_count) = np
        com_elem2D%rptr(r_count+1) =  com_elem2D%rptr(r_count)+ num_recv(np)
     end if
     if(num_send(np) /= 0) then
        s_count = s_count+1
        com_elem2D%sPE(s_count) = np
        com_elem2D%sptr(s_count+1) =  com_elem2D%sptr(s_count)+ num_send(np)
     end if
  enddo

  com_elem2D%rPEnum = r_count
  com_elem2D%sPEnum = s_count
  if (com_elem2D%rPEnum > MAX_NEIGHBOR_PARTITIONS .or.  &
       com_elem2D%sPEnum > MAX_NEIGHBOR_PARTITIONS) then
     print *,'Increase MAX_NEIGHBOR_PARTITIONS in gen_modules_partitioning.F90 and recompile'
     stop
  endif

!!$  r_count = 0
!!$  s_count = 0
!!$  allocate(com_elem2D%rptr(com_elem2D%rPEnum+1)) 
!!$  allocate(com_elem2D%sptr(com_elem2D%sPEnum+1))
  
!!$  com_elem2D%rptr(1) = 1
!!$  com_elem2D%sptr(1) = 1
!!$  
!!$  do r_count = 1, com_elem2D%rPEnum
!!$     np = com_elem2D%rPE(r_count)
!!$     com_elem2D%rptr(r_count+1) =  com_elem2D%rptr(r_count)+ num_recv(np)
!!$  enddo
!!$  do s_count = 1, com_elem2D%sPEnum
!!$     np = com_elem2D%sPE(s_count)
!!$     com_elem2D%sptr(s_count+1) =  com_elem2D%sptr(s_count)+ num_send(np)
!!$  enddo

  ! Lists themselves

  r_count = 0
  eDim_elem2D=com_elem2D%rptr(com_elem2D%rPEnum+1)-1   
  allocate(com_elem2D%rlist(eDim_elem2D))
  do np = 1,com_elem2D%rPEnum
     prank = com_elem2D%rPE(np)
     do el = 1, elem2D
        if (recv_from_pe(el) == prank) then
           r_count = r_count+1
           com_elem2D%rlist(r_count) = el
        end if
     end do
  end do
  
  s_count = 0
  allocate(com_elem2D%slist(com_elem2D%sptr(com_elem2D%sPEnum+1)-1)) 
  do np = 1,com_elem2D%sPEnum
     prank = com_elem2D%sPE(np)
     do l = 1, el_count
        el = myList_elem2D(l)
        if( any(send_to_pes(:,l) == prank)) then 
           s_count = s_count+1
           com_elem2D%slist(s_count) = el
        end if
     end do
  end do



  
  !===========================================
  !  com_elem2D_full
  !===========================================

  ! The halo relations that are already determined can be kept.
  ! Just add the elements connected only via nodes.
  ! num_send(0:npes-1) = 0
  ! num_recv(0:npes-1) = 0
  ! recv_from_pe(1:elem2D) = -1
  ! send_to_pes(1:MAX_LAENDERECK,1:elem2D) = -1
  
  ! For the local elements, collect all adjacent (sharing a node) elements
  ! that belong to other PEs

  do l=1,el_count
     el = myList_elem2D(l)

        ! Checks if triangles or quads
        q = merge(3, 4, elem2D_nodes(1,el) == elem2D_nodes(4,el))

        do n = 1, q                       ! cycle through all nodes

           nod = elem2D_nodes(n,el)

           ! Loop over all elements that belong to node nod 
           do j = 1, nod_in_elem2D_num(nod)  ! and for each node, through its patch
              elem = nod_in_elem2D(j,nod)

!!$              if (elem < 1) cycle  ! boundary, "ghost element"

              ! Check for elements to be received
              if (all(part(elem2D_nodes(1:4,elem)) /= mype) .and. recv_from_pe(elem)==-1) then
                 ! elem to be received already collected?
                 ! We have to receive elem from PE ep:
                 ep = part(elem2D_nodes(1,elem))  ! PE of first node is "main" owner
                 num_recv(ep) = num_recv(ep) + 1
                 recv_from_pe(elem) = ep  
              endif

              ! Check for elements to be sent to
              ! And maybe, we have to send el to the owners of elem
              ! This gets more complicated:
              ! 1. Is partition mype the main owner of element el?
              if (part(elem2D_nodes(1,el)) == mype .and. any(part(elem2D_nodes(1:4,elem)) /= mype)) then

                 ! 2. Who owns element elem and needs to get element el? We must check all nodes!
                 p=merge(3, 4, elem2D_nodes(1,elem) == elem2D_nodes(4,elem))

                 do i=1,p
                    ep = part(elem2D_nodes(i,elem))

                    ! 3. Is ep also an owner of el and no send is needed? This excludes also mype==ep.
                    if (any(part(elem2D_nodes(1:q,el)) == ep)) cycle

                    ! 4. Ok, for the owner ep, check if sending el is already collected
                    do k=1,MAX_LAENDERECK
                       if (send_to_pes(k,l) == ep) then
                          exit  ! already collected
                       elseif (send_to_pes(k,l) == -1) then
                          send_to_pes(k,l) = ep
                          num_send(ep) = num_send(ep) + 1
                          exit
                       elseif (k== MAX_LAENDERECK) then
                          max_laendereck_too_small = .true.  ! Problem
                       endif
                    enddo
                 end do
              endif
           end do
        end do
  enddo
  
  if (max_laendereck_too_small) then
     print *,'Increase MAX_LAENDERECK in gen_modules_partitioning.F90 and recompile'
     stop
  endif

  ! Now, build the send and recv communication data structure
!!$  com_elem2D_full%rPEnum = count(num_recv(0:npes-1) > 0)
!!$  com_elem2D_full%sPEnum = count(num_send(0:npes-1) > 0)
!!$
!!$  if (com_elem2D_full%rPEnum > MAX_NEIGHBOR_PARTITIONS .or.  &
!!$       com_elem2D_full%sPEnum > MAX_NEIGHBOR_PARTITIONS) then
!!$     print *,'Increase MAX_NEIGHBOR_PARTITIONS in gen_modules_partitioning.F90 and recompile'
!!$     stop
!!$  endif
!!$  allocate(com_elem2D_full%rPE(com_elem2D_full%rPEnum))
!!$  allocate(com_elem2D_full%sPE(com_elem2D_full%sPEnum))

!!$  r_count = 0
!!$  s_count = 0
!!$  do np = 0, npes-1
!!$     if(num_recv(np) /= 0) then
!!$        r_count = r_count+1
!!$        com_elem2D_full%rPE(r_count) = np
!!$     end if
!!$     if(num_send(np) /= 0) then
!!$        s_count = s_count+1
!!$        com_elem2D_full%sPE(s_count) = np
!!$     end if
!!$  enddo

  r_count = 0
  s_count = 0
  com_elem2D_full%rptr(1) = 1
  com_elem2D_full%sptr(1) = 1

  do np = 0, npes-1
     if(num_recv(np) /= 0) then
        r_count = r_count+1
        com_elem2D_full%rPE(r_count) = np
        com_elem2D_full%rptr(r_count+1) =  com_elem2D_full%rptr(r_count)+ num_recv(np)
     end if
     if(num_send(np) /= 0) then
        s_count = s_count+1
        com_elem2D_full%sPE(s_count) = np
        com_elem2D_full%sptr(s_count+1) =  com_elem2D_full%sptr(s_count)+ num_send(np)
     end if
  enddo

  com_elem2D_full%rPEnum = r_count
  com_elem2D_full%sPEnum = s_count

!!$  r_count = 0
!!$  s_count = 0
!!$  allocate(com_elem2D_full%rptr(com_elem2D_full%rPEnum+1)) 
!!$  allocate(com_elem2D_full%sptr(com_elem2D_full%sPEnum+1))

!!$  com_elem2D_full%rptr(1) = 1
!!$  com_elem2D_full%sptr(1) = 1
!!$
!!$  do r_count = 1, com_elem2D_full%rPEnum
!!$     np = com_elem2D_full%rPE(r_count)
!!$     com_elem2D_full%rptr(r_count+1) =  com_elem2D_full%rptr(r_count)+ num_recv(np)
!!$  enddo
!!$  do s_count = 1, com_elem2D_full%sPEnum
!!$     np = com_elem2D_full%sPE(s_count)
!!$     com_elem2D_full%sptr(s_count+1) =  com_elem2D_full%sptr(s_count)+ num_send(np)
!!$  enddo

  ! Lists themselves

  r_count = 0
  allocate(com_elem2D_full%rlist(com_elem2D_full%rptr(com_elem2D_full%rPEnum+1)-1)) 
  do np = 1,com_elem2D_full%rPEnum
     prank = com_elem2D_full%rPE(np)
     do el = 1, elem2D
        if (recv_from_pe(el) == prank) then
           r_count = r_count+1
           com_elem2D_full%rlist(r_count) = el
        end if
     end do
  end do

  s_count = 0
  allocate(com_elem2D_full%slist(com_elem2D_full%sptr(com_elem2D_full%sPEnum+1)-1)) 
  do np = 1,com_elem2D_full%sPEnum
     prank = com_elem2D_full%sPE(np)
     do l = 1, el_count
        el = myList_elem2D(l)
        if( any(send_to_pes(:,l) == prank)) then 
           s_count = s_count+1
           com_elem2D_full%slist(s_count) = el
        end if
     end do
  end do

  ! mype sends its data to
  ! comm_elem2D_full%sPEnum external PEs
  ! Their ranks (numbers) are in array
  ! comm_elem2D_full%sPE(:)
  ! Pointers to external node numbers are in
  ! comm_elem2D_full%sptr(:)
  ! The node numbers are in 
  ! comm_elem2D_full%list(:)

  deallocate(recv_from_pe, send_to_pes)

end subroutine communication_elemn



!=============================================================================


SUBROUTINE save_dist_mesh(mype)
  
  USE g_PARSUP 
  IMPLICIT NONE

  integer, intent(in) :: mype
  Integer             :: n, m, q, counter, fileID, ed(2), nglob
  character(len=10)   :: mype_string,npes_string
  character(len=200)  :: file_name
  character(len=200)  :: dist_mesh_dir
  integer, allocatable, dimension(:)  :: temp, ncount
  integer             :: edg, el, elnod(4)
INTEGER               :: g2l(max(nod2D, elem2D))

  write(mype_string,'(i5.5)') mype  
  write(npes_string,"(I10)") npes
  dist_mesh_dir=trim(meshpath)//'dist_'//trim(ADJUSTL(npes_string))//'/'

  ! ==============================
  ! rank partitioning:
  ! ==============================

  file_name=trim(dist_mesh_dir)//'my_list'//trim(mype_string)//'.out'  
  
  ! =============================   
  ! lists of owned nodes and elements
  ! =============================
  open(newunit=fileID, file=file_name)
  write(fileID,*) mype
  write(fileID,*) myDim_nod2D
  write(fileID,*) eDim_nod2D 	 
  write(fileID,*) myList_nod2D(1:myDim_nod2D), com_nod2D%rlist(1:eDim_nod2D)

  ! Create a list for eXDim_elem2D containing elements not in eDim_elem2D
  allocate(temp(com_elem2D_full%rptr(com_elem2D_full%rPEnum+1)-1))
  counter = 0
  DO n=1, com_elem2D_full%rptr(com_elem2D_full%rPEnum+1)-1
     m = com_elem2D_full%rlist(n)
     if (any(com_elem2D%rlist(1:eDim_elem2D) == m)) cycle
     counter = counter + 1
     temp(counter) = m
  end do
  eXDim_elem2D=counter

  write(fileID,*) myDim_elem2D
  write(fileID,*) eDim_elem2D
  write(fileID,*) eXDim_elem2D	 	 
  write(fileID,*) myList_elem2D(1:myDim_elem2D), com_elem2D%rlist(1:eDim_elem2D), temp(1:eXDim_elem2D)

  !NR  In TsunAWI, we set up the edges locally and don't have any halo exchange for edges
!!$  allocate(myList_edge2D(4*myDim_elem2D))
!!$  counter = 0
!!$  do edg=1, edge2D
!!$     do q=1,2 
!!$        if (part(edges(q,edg))==mype) then
!!$           counter=counter+1
!!$           myList_edge2D(counter)=edg
!!$           exit
!!$        end if
!!$     end do
!!$  end do
!!$  myDim_edge2D=counter   ! It is the total number of my edges
!!$  
!!$  Do el=1,myDim_elem2D
!!$
!!$
!!$     DO q=1,4
!!$        edg = elem_edges(q,myList_elem2D(el))
!!$        
!!$        if (edg > 0) then  ! edg=0 for q=4 on triangles
!!$           if ((       part(edges(1,edg)) /= mype) &
!!$                .and. (part(edges(2,edg)) /= mype) &
!!$                .and. all(myList_edge2D(myDim_edge2D:counter) /= edg)) then
!!$              counter = counter+1 
!!$              myList_edge2D(counter) = edg
!!$           end if
!!$        endif
!!$     END DO
!!$  END DO
!!$  eDim_edge2D = counter-myDim_edge2D
!!$  if (myDim_edge2D+eDim_edge2D > 4*myDim_elem2D) then
!!$     write (*,*) 'Edge array too short in save_dist_mesh!'
!!$     stop
!!$  endif
!!$
!!$  write(fileID,*) myDim_edge2D
!!$  write(fileID,*) eDim_edge2D 	 
!!$  write(fileID,*) myList_edge2D(1:myDim_edge2D +eDim_edge2D)
!!$  deallocate(myList_edge2D)

  close(fileID)       

  
  !================================
  ! From global to local numbering
  !================================

  
! ===============
! nodes: Local numbering
! ===============
  ! Replace global numbering with a local one
  g2l(:)=0
  DO n=1, myDim_nod2D
     g2l(myList_nod2D(n)) = n
  END DO

  DO n=1, eDim_nod2D
     g2l(com_nod2D%rlist(n)) = myDim_nod2D+n 
     com_nod2D%rlist(n)      = myDim_nod2D+n
  END DO

  DO n=1, com_nod2D%sptr(com_nod2D%sPEnum+1)-1
     com_nod2D%slist(n) = g2l(com_nod2D%slist(n))
  END DO

!============================================
! local element definition: map node numbers 
!============================================
  !
  ! Remark: Most routines already account for quads.
  ! Here, we only write 3 columns, to not mess with TsunAWI
  
  file_name=trim(dist_mesh_dir)//'elem2d_'//trim(mype_string)//'.out' 
    
  open(newunit=fileID, file=file_name)

  ! Inner elements, all nodes are known to this PE
  do el=1, myDim_elem2D 
     write(fileID,*) g2l(elem2D_nodes(1:3, myList_elem2D(el)))
  end do

  ! Halo elements, the outer rise nodes are not located on this PE 
  do m = 1, eDim_elem2D
     el = myDim_elem2D + m
     do q=1,3
        nglob = elem2D_nodes(q, com_elem2D%rlist(m))
        if (g2l(nglob) > 0) then
           elnod(q) = g2l(nglob)
        else
           elnod(q) = -nglob  ! mype does not own this node at all
        end if
     end do
     write(fileID,*) elnod(1:3)
  end do
  
  do m = 1, exDim_elem2D
     ! temp contains only those elements in exDim_elem2D that do not
     ! belong to eDim_elem2D
     el = temp(m) 
     do q=1,3
        nglob = elem2D_nodes(q, el)
        if (g2l(nglob) > 0) then
           elnod(q) = g2l(nglob)
        else
           elnod(q) = -nglob
        end if
     end do
     write(fileID,*) elnod(1:3)
  end do
  close(fileID)

  deallocate(temp)
  
! =================
! element numbering
! =================
  ! Replace global numbering with a local one
  g2l(:)=0
  DO el=1, myDim_elem2D
     g2l(myList_elem2D(el))=el
  END DO

  DO el=1, eDim_elem2D
     g2l(com_elem2D%rlist(el)) = myDim_elem2D+el
     com_elem2D%rlist(el)      = myDim_elem2D+el
  END DO

  DO el=1, com_elem2D%sptr(com_elem2D%sPEnum+1)-1
     com_elem2D%slist(el)=g2l(com_elem2D%slist(el))
  END DO


! =========
! elements (extra needed for MUSCL advection)
! =========
  
  m = 0
  DO el=1, com_elem2D_full%rptr(com_elem2D_full%rPEnum+1)-1
     if (g2l(com_elem2D_full%rlist(el)) > 0) then
        com_elem2D_full%rlist(el) = g2l(com_elem2D_full%rlist(el))
     else
        m = m+1
        com_elem2D_full%rlist(el) = myDim_elem2D+eDim_elem2D+m
     endif
  END DO
  
  DO el=1, com_elem2D_full%sptr(com_elem2D_full%sPEnum+1)-1
     com_elem2D_full%slist(el) = g2l(com_elem2D_full%slist(el))
  END DO

  
  

  ! =========================  
  ! communication information
  ! ========================= 
  file_name=trim(dist_mesh_dir)//'com_info'//trim(mype_string)//'.out' 
    
  open(newunit=fileID, file=file_name)
  write(fileID,*) mype
  write(fileID,*) com_nod2D%rPEnum
  write(fileID,*) com_nod2D%rPE(1:com_nod2D%rPEnum)
  write(fileID,*) com_nod2D%rptr(1:com_nod2D%rPEnum+1)
  write(fileID,*) com_nod2D%rlist
  write(fileID,*) com_nod2D%sPEnum
  write(fileID,*) com_nod2D%sPE(1:com_nod2D%sPEnum)
  write(fileID,*) com_nod2D%sptr(1:com_nod2D%sPEnum+1)
  write(fileID,*) com_nod2D%slist
  deallocate(myList_nod2D)
!!$  deallocate(com_nod2D%rPE)
!!$  deallocate(com_nod2D%rptr)
  deallocate(com_nod2D%rlist)
!!$  deallocate(com_nod2D%sPE)
!!$  deallocate(com_nod2D%sptr)
  deallocate(com_nod2D%slist)

  write(fileID,*) com_elem2D%rPEnum
  write(fileID,*) com_elem2D%rPE(1:com_elem2D%rPEnum)
  write(fileID,*) com_elem2D%rptr(1:com_elem2D%rPEnum+1)
  write(fileID,*) com_elem2D%rlist
  write(fileID,*) com_elem2D%sPEnum
  write(fileID,*) com_elem2D%sPE(1:com_elem2D%sPEnum)
  write(fileID,*) com_elem2D%sptr(1:com_elem2D%sPEnum+1)
  write(fileID,*) com_elem2D%slist
  deallocate(myList_elem2D)
!!$  deallocate(com_elem2D%rPE)
!!$  deallocate(com_elem2D%rptr)
  deallocate(com_elem2D%rlist)
!!$  deallocate(com_elem2D%sPE)
!!$  deallocate(com_elem2D%sptr)
  deallocate(com_elem2D%slist)

  write(fileID,*) com_elem2D_full%rPEnum
  write(fileID,*) com_elem2D_full%rPE(1:com_elem2D_full%rPEnum)
  write(fileID,*) com_elem2D_full%rptr(1:com_elem2D_full%rPEnum+1)
  write(fileID,*) com_elem2D_full%rlist
  write(fileID,*) com_elem2D_full%sPEnum
  write(fileID,*) com_elem2D_full%sPE(1:com_elem2D_full%sPEnum)
  write(fileID,*) com_elem2D_full%sptr(1:com_elem2D_full%sPEnum+1)
  write(fileID,*) com_elem2D_full%slist
!!$  deallocate(com_elem2D_full%rPE)
!!$  deallocate(com_elem2D_full%rptr)
  deallocate(com_elem2D_full%rlist)
!!$  deallocate(com_elem2D_full%sPE)
!!$  deallocate(com_elem2D_full%sptr)
  deallocate(com_elem2D_full%slist)
  close(fileID)
  ! ================================
  ! mapping ( PE contiguous 2D numbering) 	 
  ! ================================  
  if(mype==0) then
     file_name=trim(dist_mesh_dir)//'rpart.out'  
     open(newunit=fileID, file=file_name)

     allocate(temp(nod2D), ncount(npes+1)) ! mapping 
     ncount=0
     DO n=1, nod2D
        m=part(n)
        ncount(m+1)=ncount(m+1)+1
     END DO
     write(fileID,*) npes
     write(fileID,*) ncount(1:npes)

     ncount=0
     DO n=1, nod2D
        m=part(n)
        ncount(m+2)=ncount(m+2)+1
        temp(n)=ncount(m+2)
     END DO
     ncount(1)=1
     Do n=2,npes+1	  
        ncount(n)=ncount(n)+ncount(n-1)
     end do
     ! Now count == part in range partitioning   

     do n=1,nod2D
        temp(n)=temp(n)+ncount(part(n)+1)-1
        write(fileID,*) temp(n)  
     end do

     deallocate(ncount, temp)
     close(fileID)

     file_name=trim(dist_mesh_dir)//'part.out'  
     open(newunit=fileID, file=file_name)
     do n=1,nod2D
        write(fileID,*) part(n)  
     end do
     close(fileID)
  end if

  

!!$  write(*,*) 'Distributed mesh is saved for rank ', mype
END subroutine  save_dist_mesh
