! Part of TsunAWI
! Copyright (c) 2021
! Alfred Wegener Institut,
! Helmholtz Centre for Polar and Marine Research
! Contact: rakowsky@awi.de


! !MODULE: OKADA_FAULT_PARAMETERS - contains fault parameters andvariables nee-
!	ded for the determination of initial conditions from these parameters
! !DESCRIPTION: This module contains parameters related to initial conditions 
!               based on Okada parameters. These are read from a file with all
!               sets of parameters. Starting from these parameters the 
!               initial conditions are determined 
!               by a tool developed in the IUGG/IOC TIME Project.
!
! Definitions are:\\
! \begin{tabular}{l|lr}\hline
! Name & Description                    & Unit \\ \hline
! pi   & $\pi=3.141592653589793$        & -- \\
! rad  & radians: $rad=\frac{\pi}{180}$ & -- \\
! X0   & Fault origin longitude         & deg west \\
! Y0   & Fault Origin latitude          & deg north\\
! D    & Slip (Dislocation) & m\\
! L    & fault length & m \\
! W    & fault width & m\\
! TH   & Strike Angle & deg \\
! DL   & Dip Angle & deg\\
! HH   & fault depth & m\\
! RD   & Direction of dislocation & deg\\
! XCD  & origin of computational domain lon & deg west \\
! YCD  & origin of computational domain lat & deg north\\
! nfault & number of fault planes in simulation& -- \\
! ttl\_nfault & total number of fault planes & -- \\
! okada_rupt  & array with list of faults in simulation (size: nfault) & -- \\
! AllFaults & array with all fault parameters (size: ttl\_nfault x 11) & -- \\ \hline
! \end{tabular}


MODULE OKADA_FAULT_PARAMETERS

    use PARAMETERS
    use MPI_COMMUNICATION
    IMPLICIT NONE
    private

    type t_okada
       real(kind=dp) :: lon     ! degree
       real(kind=dp) :: lat     ! degree
       real(kind=dp) :: slip    ! [m]
       real(kind=dp) :: len     ! [m]
       real(kind=dp) :: width   ! [m]
       real(kind=dp) :: strike  ! degree
       real(kind=dp) :: dip     ! degree 
       real(kind=dp) :: depth   ! of top [m]
       real(kind=dp) :: rake    ! degree
       real(kind=dp) :: lon_SW  ! degree \ Bounding box lower left
       real(kind=dp) :: lat_SW  ! degree / of interpolation area
       character     :: ref_point ! 'o' = Origin (SW corner), 'c' = Center   
    end type t_okada

    public :: t_okada, okada_read_faultlist, okada_update
    public :: nfault, okada_faults, okada_rupt
    SAVE
    
    INTEGER                                    :: nfault

    ! Dimension of region in which to compute the displacement for each fault.
    ! For historical reasons, it is a hardcoded square of 8 Degree x 8 Degree.
    real(kind=dp), parameter                   :: side_lat=8., side_lon=8.
    REAL(KIND=dp), PARAMETER                   :: Z_thresh = 0.000001  !0.30 0.05
  
    !  List of faults from file
    INTEGER,      ALLOCATABLE, DIMENSION(:)    :: okada_rupt 
    TYPE(t_okada), ALLOCATABLE, DIMENSION(:)   :: okada_faults

    integer, allocatable                       :: i_rupttime(:)
    integer                                    :: i_okada_last_rupt_time
CONTAINS

 SUBROUTINE okada_read_faultlist

   use mpi_parallel
   
  implicit none

  CHARACTER(len=10)          :: okada_format
  CHARACTER(len=100)         :: okada_parameter_file, okada_fault_list_file
  INTEGER                    :: infile_nml, stat 

  NAMELIST /model_init_okada_details/ okada_format, okada_parameter_file, okada_fault_list_file

  ! Read the namelist section 
  okada_format=""
  okada_parameter_file= ""
  okada_fault_list_file=""

  OPEN(newunit=infile_nml, file = 'namelist.tsunami', iostat=stat)
  read(infile_nml, NML  = model_init_okada_details, iostat=stat)
  CLOSE(infile_nml)
  if (stat /= 0 .and. mype==0) then
     print *,'/model_init_okada_details/ is missing or corrupt in namelist.tsunami'
     print *,'                           Falling back to defaults.'
  end if

  if (len_trim(okada_parameter_file)==0) then
     if (mype==0) print *,"/model_init_okada_details/ Setting missing okada_parameter_file = 'okada_parameter_file'"
     okada_parameter_file = 'okada_parameter_file'
  endif

  if (len_trim(okada_fault_list_file)==0) then
     if (mype==0) print *,"/model_init_okada_details/ Setting missing okada_fault_list_file = 'fault_list'"
     okada_fault_list_file = 'fault_list'
  endif


  if (len_trim(okada_format)==0) then
     if (index(trim(adjustl(okada_parameter_file)),'.param',back=.true.) == len_trim(adjustl(okada_parameter_file))-5) then
        ! The suffix is .param and we assume USGS param format
        okada_format='usgs.param'
     else
        okada_format='tsunawi'
     endif
     if (mype==0) then
        print *,"/model_init_okada_details/ Setting missing okada_format = '"//trim(adjustl(okada_format))//"'"
        print *,'                           (determined from suffix of okada_parameter_file)'
     end if
  endif

  ! Now, read the Okada parameters
  
  if (trim(adjustl(okada_format))=='usgs.param') then
     call okada_read_usgs_faultlist(okada_parameter_file)
  elseif (trim(adjustl(okada_format))=='tsunawi') then
     call okada_read_tsunawi_faultlist(okada_parameter_file, okada_fault_list_file)
  else
     if (mype==0) print *,"/model_init_okada_details/ Sorry, okada_format = '"//trim(adjustl(okada_format))//"' is not implemented."
     stop
  end if

end SUBROUTINE okada_read_faultlist

subroutine okada_read_tsunawi_faultlist(okada_parameter_file, okada_fault_list_file)

  use mpi_parallel
  
  character(len=*), intent(in) :: okada_parameter_file, okada_fault_list_file
  integer                      :: iostatus, open_status, n, ttl_nfault, n_comment
  real(kind=dp)                :: values(11)
  character(len=200)           :: line
  character(len=1)             :: c1, ref_point
  real(kind=dp), allocatable   :: okada_rupt_delay(:) 
  INTEGER                      :: infile_okada, infile_list, infile_nml 

  ! Determine number of lines = number of faults
  open(newunit=infile_okada, file = okada_parameter_file, form='formatted', action='read')
  iostatus   = 0
  ttl_nfault = 0
  n_comment  = 0
  do while(iostatus==0)
     read (infile_okada,*,iostat=iostatus) line
     line = adjustl(line)
     c1=line(1:1)
     if (c1=='#'  .or. c1=='%' .or. c1=='!' .or. &
         c1=='c'  .or. c1=='C' .or. c1=='o' .or. c1=='O') then
        n_comment=n_comment+1
     else
         ttl_nfault = ttl_nfault+1
     endif
  end do
  close(infile_okada)

  ttl_nfault=ttl_nfault-1
  if (mype==0) print *,'Total number of faults in file: ',ttl_nfault
  
  allocate(okada_faults(ttl_nfault))
  
  OPEN(newunit=infile_okada, file=okada_parameter_file, form='formatted', action='read')
  n = 0
  ref_point = 'o'
  DO while (n < ttl_nfault)
     read(infile_okada,'(A)',iostat=iostatus) line
     !line = adjustl(line)
     c1=line(1:1)
     
     if (c1=='#' .or. c1=='%' .or. c1=='!') then
        cycle ! comment
     elseif (c1=='c' .or. c1=='C') then
        ref_point='c'    ! The plate coordinates are given at the centers of the plates
                         ! The default is "at the origin"
        cycle
     endif

     read(line,*,iostat=iostatus) values(1:11)
     if (iostatus /= 0) then
        ! The corner of the bounding box (XCD, YCD) is not specified
        read(line,*,iostat=iostatus) values(1:9)
        ! substract 8 degrees, as the interpolation mesh is 16 degr x 16 degr 
        ! (currently, side_lon, lat = 8. as hard coded parameter - it has never been
        !  changed so far...)
        values(10) = values(1) - .5*side_lon
        values(11) = values(2) - .5*side_lat
     endif
     
     n = n+1

     okada_faults(n)%lon    = values(1) 
     okada_faults(n)%lat    = values(2) 
     okada_faults(n)%slip   = values(3) 
     okada_faults(n)%len    = values(4) 
     okada_faults(n)%width  = values(5) 
     okada_faults(n)%strike = values(6) 
     okada_faults(n)%dip    = values(7) 
     okada_faults(n)%depth  = values(8) 
     okada_faults(n)%rake   = values(9) 
     okada_faults(n)%lon_SW = values(10)
     okada_faults(n)%lat_SW = values(11)

  END DO
  close(infile_okada)

  if (ref_point=='c') then
     do n=1, ttl_nfault
        ! If required, transform plate parameters from center to origin
        okada_faults(n)%ref_point = ref_point
        if (n==1) print *,'Transforming coordinates from center to origin of fault'
        call center2origin(okada_faults(n))
     enddo
  end if
  
  open(newunit=infile_list, file = okada_fault_list_file, form='formatted', action='read')
  read(infile_list, *) nfault
  allocate(okada_rupt(nfault))
  allocate(okada_rupt_delay(nfault))
  allocate(i_rupttime(nfault))
  
  read(infile_list, *) okada_rupt(:)
  read(infile_list, *) okada_rupt_delay(:)

  i_rupttime(:) = INT(okada_rupt_delay(:)/dt)
  
  i_okada_last_rupt_time = maxval(i_rupttime(1:nfault))
  close(infile_list)

  deallocate(okada_rupt_delay)
  
END SUBROUTINE okada_read_tsunawi_faultlist
!====

subroutine okada_read_usgs_faultlist(okada_parameter_file)

  character(len=*), intent(in) :: okada_parameter_file

  ! Final Faults from USGS provide the same values as TsunAWI's format, just in a different order
  ! This routine reads the old "param" format, see 
  ! https://earthquake.usgs.gov/data/finitefault/

  integer                    :: infile_usgs
  character(len=1)           :: c_dummy
  character(len=200)         :: substring, faultsegment

  integer                    :: iA, len
  integer                    :: nx, ny, n
  real(kind=sp)              :: Dx, Dy, values(11)
  real(kind=dp)              :: okada_rupt_delay
  
  open(newunit=infile_usgs, file=okada_parameter_file, form='formatted', action='read' )

  read(infile_usgs,*) c_dummy
  read(infile_usgs,'(A)') faultsegment

  ! extract nx, ny => nx*ny plates
  ! extract Dx, Dy => each plate has size Dx * Dy [km*km] 

   iA = index(faultsegment,'nx(Along-strike)=')
   len = index(faultsegment(iA:),'Dx')
   substring = faultsegment(iA+17:iA+len-2)
   read(substring,*) nx

   iA = index(faultsegment,'Dx=')
   len = index(faultsegment(iA:),'km')
   substring = faultsegment(iA+3:iA+len-2)
   read(substring,*) Dx

   iA = index(faultsegment,'ny(downdip)=')
   len = index(faultsegment(iA:),'Dy')
   substring = faultsegment(iA+12:iA+len-2)
   read(substring,*) ny

   iA = index(faultsegment,'Dy=')
   len = index(faultsegment(iA:),'km')
   substring = faultsegment(iA+3:iA+len-2)
   read(substring,*) Dy

   ! Another 8 lines with redundant information 
   do n=1,8
      read(infile_usgs,*) c_dummy
   end do
  
   ! allocate the arrays for faults and timeline
   nfault = nx*ny
   allocate(okada_faults(nfault))
   allocate(okada_rupt(nfault))
   allocate(i_rupttime(nfault))

   ! Read the values from the USGS parameter file
   ! and resort into our t_okada structure
   do n=1,nfault
      read(infile_usgs,*) values(1:11)
      !    1    2      3    4    5      6   7     8     9    10 11
      ! #Lat. Lon. depth slip rake strike dip t_rup t_ris t_fal mo
      okada_faults(n)%lon    = values(2) 
      okada_faults(n)%lat    = values(1) 
      okada_faults(n)%slip   = values(4)*.01   ! scale from cm to m
      okada_faults(n)%len    = Dx*1000.        ! scale from km to m
      okada_faults(n)%width  = Dy*1000.        ! scale from km to m
      okada_faults(n)%strike = values(6) 
      okada_faults(n)%dip    = values(7) 
      okada_faults(n)%depth  = values(3)*1000. ! scale from km to m
      okada_faults(n)%rake   = values(5) 
      okada_faults(n)%lon_SW = okada_faults(n)%lon - .5*side_lon
      okada_faults(n)%lat_SW = okada_faults(n)%lat  -.5*side_lat     
      okada_faults(n)%ref_point = 'c'  ! USGS param gives center of plate

      okada_rupt(n)          = n
      okada_rupt_delay       = values(8)

      i_rupttime(:) = INT(okada_rupt_delay/dt)

      call center2origin(okada_faults(n))
   end do
  
  i_okada_last_rupt_time = maxval(i_rupttime(1:nfault))

   close(infile_usgs)
   
end subroutine okada_read_usgs_faultlist

!===

SUBROUTINE okada_update(i_time)
!  Update ssh if the rupture takes several time steps
  use mesh
  use swe
  
  IMPLICIT NONE
  INTEGER, intent(in) :: i_time

  INTEGER       :: n
  LOGICAL       :: l_update = .false.
    
  DO n = 1 , nfault

     IF (i_time == i_rupttime(n)) THEN
        !$OMP MASTER
        if (mype==0) PRINT *,'Processing Fault Nmb. ',n,okada_rupt(n)
        !$OMP END MASTER
        
        CALL deform(okada_faults(okada_rupt(n)))
        l_update = .true.
     END IF
  END DO
  
  if (l_update) then
!$OMP BARRIER
!$OMP MASTER
     call com_halo_begin(com_nod2D, ssh0)
     call com_halo_begin(com_nod2D, ssh1)
     call com_halo_begin(com_nod2D, ssh2)
     call com_halo_begin(com_nod2D, nodhn)
     call com_halo_end(  com_nod2D)
!$OMP END MASTER
!$OMP BARRIER
  endif
  
END SUBROUTINE okada_update


SUBROUTINE center2origin(fault)

    !Transform coordinates from center of the fault plate to the origin
    implicit none

    type(t_okada), intent(inout) :: fault

    REAL(KIND=dp)                :: Lh, Wh, cos_strike, sin_strike

    if (fault%ref_point == 'c') then
       cos_strike = cos(fault%strike * rad)
       sin_strike = sin(fault%strike * rad)
       
       ! half the len and width in degrees       
       Lh = 0.5*fault%len                         * rad_inv*r_earth_inv  
       Wh = 0.5*fault%width *cos(fault%dip * rad) * rad_inv*r_earth_inv  
       
       ! -st because Strike angle is measured in clockwise direction
       fault%lon       = fault%lon - Wh * cos_strike - Lh * sin_strike 
       fault%lat       = fault%lat + Wh * sin_strike - Lh * cos_strike
       fault%ref_point = 'o'
    endif

  END SUBROUTINE center2origin


REAL(KIND=dp) FUNCTION atn(ax, ay)

    IMPLICIT NONE

    REAL(KIND=dp)            :: ax, ay

    REAL(KIND=dp), PARAMETER :: gx = 1.0e-6
    REAL(KIND=dp)            :: aax, aay, p, sr

    aax = ABS(ax) ; aay = ABS(ay)
    
  
    IF (aax <= gx .AND. aay <= gx) THEN
    !    WRITE(6,'("ATAN -- ax=",E15.7,2X,"ay=",E15.7)') ax,ay
        atn = 0.2
    ELSE
        atn = sign(atan2(aax,aay),ax*ay)
    END IF

END FUNCTION atn


SUBROUTINE deform(fault)

    USE MESH
    USE SWE
        
    IMPLICIT NONE

    type(t_okada), intent(in) :: fault 

    REAL(kind=wp), save      :: dssh_max, dssh_min
    REAL(kind=dp), parameter :: inv_12pi = 1./(12.*pi) 
    INTEGER        :: n
    REAL(KIND=wp)  :: dssh
    REAL(KIND=dp) :: XL, YL, H1, H2, DS, DD
    REAL(KIND=dp) :: XX, YY, X1, X2 
    REAL(KIND=dp) :: F1, F2, F3, F4, G1, G2, G3, G4
    REAL(KIND=dp) :: sin_TH, cos_TH, HH_over_tanDL
    REAL(KIND=dp) :: sin_DL, cos_DL, tan_DL
    REAL(kind=dp) :: L, XCD, YCD
    
    L   = 0.5*Fault%len
    XCD = rad*fault%lon_SW ! SW corner of covered domain
    YCD = rad*fault%lat_SW 
    
    sin_DL = sin(rad*Fault%dip)
    cos_DL = cos(rad*Fault%dip)
    tan_DL = tan(rad*Fault%dip)

    H1 = Fault%depth / sin_DL
    H2 = H1 + Fault%width  
    DS =  Fault%slip * cos(rad*Fault%rake)*inv_12pi
    DD =  Fault%slip * sin(rad*Fault%rake)*inv_12pi

    sin_TH = sin(rad*Fault%strike)
    cos_TH = cos(rad*Fault%strike)
    HH_over_tanDL = Fault%depth /tan_DL

    !$OMP MASTER
    dssh_max = -1000.
    dssh_min = +1000.
    !$OMP END MASTER
    !$OMP BARRIER

    !$OMP DO REDUCTION(max:dssh_max) REDUCTION(min:dssh_min)
    DO n = 1,myDim_nod2D

       IF (coord_nod(1,n) > XCD .AND. coord_nod(1,n) < XCD + rad*side_lon  .AND.      &
           coord_nod(2,n) > YCD .AND. coord_nod(2,n) < YCD + rad*side_lat) THEN

           ! cartesian relative to Fault%lon, Fault%lat
          XX    = r_earth*(coord_nod(1,n) - rad*Fault%lon) * cos_nod(n)
          YY    = r_earth*(coord_nod(2,n) - rad*Fault%lat)
          
          X1 = XX * sin_TH + YY*cos_TH - L
          X2 = XX * cos_TH - YY*sin_TH + HH_over_tanDL
        
          CALL ucal(X1, X2,  L, H2, sin_DL, cos_DL, tan_DL, F1, G1)
          CALL ucal(X1, X2,  L, H1, sin_DL, cos_DL, tan_DL, F2, G2)
          CALL ucal(X1, X2, -L, H2, sin_DL, cos_DL, tan_DL, F3, G3)
          CALL ucal(X1, X2, -L, H1, sin_DL, cos_DL, tan_DL, F4, G4)

          dssh = (F1-F2-F3+F4)*DS + (G1-G2-G3+G4)*DD
          
         IF(ABS(dssh) <= Z_thresh) dssh = 0._wp
          
          ssh0(n) = ssh0(n) + dssh
          ssh1(n) = ssh1(n) + dssh
          ssh2(n) = ssh2(n) + dssh

          ssh_init(n) = ssh_init(n)+dssh

          nodhn(n) = nodhn(n) - dssh

          dssh_max = max(dssh_max, dssh)
          dssh_min = min(dssh_min, dssh)
       END IF

    END DO
    !$OMP END DO

    call com_reduce_begin(dssh_max, com_max)
    call com_reduce_begin(dssh_min, com_min)
    call com_reduce_end

    !$OMP MASTER
    if (mype == 0) print *,'Displacement min, max', dssh_min, dssh_max
    !$OMP END MASTER
    
  END SUBROUTINE deform

SUBROUTINE ucal(X1, X2, C, CC, sn, cs, tn, F, G)

    IMPLICIT NONE

    REAL(KIND=dp), INTENT(IN)  :: X1, X2, C, CC, sn, cs, tn

    REAL(KIND=dp), INTENT(OUT) :: F, G

    REAL(KIND=dp) :: c2, c3
    REAL(KIND=dp) :: q, q2, q3 
    REAL(KIND=dp) :: h, k
    REAL(KIND=dp) :: a1, a2, a3
    REAL(KIND=dp) :: b1, b2, b3, b4, b5, b6, b8, b9
    REAL(KIND=dp) :: bb1, d1, d2, d5    

    ! computation of the vertical displacement due to strike slip
    
    c2 = CC*cs
    c3 = CC*sn
    q   = sqrt((X1-C)**2+(X2-c2)**2+c3**2)
    q2  = X2*sn
    q3  =-X2*cs
    a1  = log(q-q3-CC)
    a2  = log(q+q3+CC)
    a3  = log(q+c3)
    b1  = 1. + 3.0*tn*tn
    b2  = 3.0 * tn/cs
    b3  = 2.0 * q2 * sn
    b4  = q2 + x2*sn
    b5  = 2.0 * q2*q2*cs
    b6  = q*(q-q3-CC)
    b8  = 2.0 * (q2+X2*sn) * (q3*sn)
    b9  = q*(q+q3+CC)
    
    F   = cs * (a1 + b1*a2 -b2*a3)                  &
          + (b3 + 2.0*sn*b4)/q - b5/b6              &
          + b8/b9 

! computation of the vertical displacement due to dip slip

    h  = sqrt(q2**2+(q3+CC)**2)
    k  = sqrt((X1-C)**2+q2**2)
    bb1 = 1./(q*(q+x1-C))
    d1 = X1-C
    d2 = X2-c2
    d5 = -q3-CC 
  
    G  = sn*(-d2*6.*c3*bb1      &
         - 6.*atn(d1*d2, (h+c3)*(q+k)) + 9.*atn(d1*d5, q2*q))      &
         + cs * (-6.*c3*c3*bb1) 

  END SUBROUTINE ucal

  
END MODULE OKADA_FAULT_PARAMETERS

!NR  Keep for reference: The original version of uscal, udcal,
!NR  w/o optimizations for X3=0. and precomputed sin, cos, tan


!!$SUBROUTINE uscal(X1, X2, X3, C, CC, DPL, F)
!!$! computation of the vertical displacement due to strike slip
!!$
!!$    IMPLICIT NONE
!!$
!!$    REAL(KIND=dp), INTENT(IN)  :: X1, X2, X3, C, CC, DPL
!!$    REAL(KIND=dp), INTENT(OUT) :: F
!!$
!!$    REAL(KIND=dp) :: sn, cs, c1, c2, c3
!!$    REAL(KIND=dp) :: r, r2, r3, q, q2, q3 
!!$    REAL(KIND=dp) :: h, k
!!$    REAL(KIND=dp) :: a1, a2, a3
!!$    REAL(KIND=dp) :: b1, b2, b3, b4, b5, b6, b7, b8, b9
!!$    REAL(KIND=dp) :: b10, b11, b12, b13, b14
!!$
!!$    sn  = sin(DPL)
!!$    cs  = cos(DPL)
!!$    c1  = C; c2 = CC*cs; c3 = CC*sn
!!$    r   = sqrt((X1-c1)**2+(X2-c2)**2+(X3-c3)**2)
!!$    q   = sqrt((X1-c1)**2+(X2-c2)**2+(X3+c3)**2)
!!$    r2  = X2*sn - X3*cs
!!$    r3  = X2*cs - X3*sn
!!$    q2  = X2*sn + X3*cs
!!$    q3  =-X2*cs + X3*sn
!!$    h   = sqrt(q2**2+(q3+CC)**2)
!!$    k   = sqrt((X1-c1)**2+q2**2)
!!$    a1  = log(r+r3-CC)
!!$    a2  = log(q+q3+CC)
!!$    a3  = log(q+X3+c3)
!!$    b1  = 1. + 3.0*(tan(DPL))**2
!!$    b2  = 3.0 * tan(DPL)/cs
!!$    b3  = 2.0 * r2 * sn
!!$    b4  = q2 + x2*sn
!!$    b5  = 2.0 * r2**2*cs
!!$    b6  = r*(r+r3-CC)
!!$    b7  = 4.0 * q2 * X3 * sn**2
!!$    b8  = 2.0 * (q2+X2*sn) * (x3+q3*sn)
!!$    b9  = q * (q+q3+CC)
!!$    b10 = 4.0 * q2 * x3 * sn
!!$    b11 = (x3+c3) - q3*sn
!!$    b12 = 4.0 * q2**2 * q3*x3*cs*sn
!!$    b13 = 2.0*q+q3+cc
!!$    b14 = q**3 * (q+q3+CC)
!!$
!!$    F   = cs * (a1 + b1*a2 -b2*a3)                  &
!!$          + b3/r + 2.0*sn*b4/q - b5/b6              &
!!$          + (b7-b8)/b9 + b10*b11/q**3 - b12*b13/b14
!!$
!!$END SUBROUTINE uscal
!!$
!!$
!!$SUBROUTINE udcal(X1, X2, X3, C, CC, DPL, F)
!!$! computation of the vertical displacement due to dip slip
!!$
!!$    IMPLICIT NONE
!!$
!!$    REAL(KIND=dp), INTENT(IN)  :: X1, X2, X3, C, CC, DPL
!!$    REAL(KIND=dp), INTENT(OUT) :: F
!!$
!!$    REAL(KIND=dp) :: sn, cs, c1, c2, c3
!!$    REAL(KIND=dp) :: r, r2, r3, q, q2, q3 
!!$    REAL(KIND=dp) :: h, k
!!$    REAL(KIND=dp) :: a1, a2
!!$    REAL(KIND=dp) :: b1, b2, b3, d1, d2, d3, d4, d5, d6
!!$    REAL(KIND=dp) :: t1, t2, t3
!!$!    REAL(KIND=dp), EXTERNAL :: atn
!!$
!!$    sn = sin(DPL)
!!$    cs = cos(DPL)
!!$    c1 = C; c2 = CC*cs; c3 = CC*sn
!!$    r  = sqrt((X1-c1)**2+(X2-c2)**2+(X3-c3)**2)
!!$    q  = sqrt((X1-c1)**2+(X2-c2)**2+(X3+c3)**2)
!!$    r2 = X2*sn - X3*cs
!!$    r3 = X2*cs - X3*sn
!!$    q2 = X2*sn + X3*cs
!!$    q3 =-X2*cs + X3*sn
!!$    h  = sqrt(q2**2+(q3+CC)**2)
!!$    k  = sqrt((X1-c1)**2+q2**2)
!!$    a1 = log(r+X1-c1)
!!$    a2 = log(q+X1-c1)
!!$    b1 = q*(q+x1-c1)
!!$    b2 = r*(r+x1-c1)
!!$    b3 = q*(q+q3+CC)
!!$    d1 = X1-c1 ; d2 = X2-c2 ; d3 = X3-c3
!!$    d4 = X3+c3 ; d5 = r3-CC ; d6 = q3+CC
!!$    t1 = atn(d1*d2, (h+d4)*(q+k))
!!$    t2 = atn(d1*d5, r2*r)
!!$    t3 = atn(d1*d6, q2*q)
!!$  
!!$    F  = sn*(d2*(2*d3/b2+4*d3/b1-4*c3*X3*d4*(2*q+d1)/(b1**2*q))      &
!!$         - 6*t1 + 3*t2 -6*t3)                                        &
!!$         + cs * (a1-a2-2*(d3**2)/b2                                  &
!!$         - 4*(d4**2-c3*X3)/b1 - 4*c3*X3*d4**2*(2*q+X1-c1)/(b1**2*q)) &
!!$         + 6*X3 * (cs*sn*(2*d6/b1+d1/b3)-q2*(sn**2-cs**2)/b1)
!!$
!!$END SUBROUTINE udcal
